"""
NANP Area Code to State/Region/Province Mapping
Comprehensive mapping of area codes to their geographic locations.

Source: North American Numbering Plan Administrator (NANPA)
Last Updated: January 2025
"""

# Area code to state/region/province mapping
AREA_CODE_MAPPING = {
    # 200s
    201: {'state': 'NJ', 'region': 'North Jersey', 'country': 'US'},
    202: {'state': 'DC', 'region': 'Washington, D.C.', 'country': 'US'},
    203: {'state': 'CT', 'region': 'Southwest Connecticut', 'country': 'US'},
    204: {'state': 'MB', 'region': 'Manitoba', 'country': 'CA'},
    205: {'state': 'AL', 'region': 'Central/West Alabama', 'country': 'US'},
    206: {'state': 'WA', 'region': 'Seattle', 'country': 'US'},
    207: {'state': 'ME', 'region': 'Maine', 'country': 'US'},
    208: {'state': 'ID', 'region': 'Idaho', 'country': 'US'},
    209: {'state': 'CA', 'region': 'Central California', 'country': 'US'},
    210: {'state': 'TX', 'region': 'San Antonio', 'country': 'US'},
    212: {'state': 'NY', 'region': 'Manhattan', 'country': 'US'},
    213: {'state': 'CA', 'region': 'Los Angeles', 'country': 'US'},
    214: {'state': 'TX', 'region': 'Dallas', 'country': 'US'},
    215: {'state': 'PA', 'region': 'Philadelphia', 'country': 'US'},
    216: {'state': 'OH', 'region': 'Cleveland', 'country': 'US'},
    217: {'state': 'IL', 'region': 'Central Illinois', 'country': 'US'},
    218: {'state': 'MN', 'region': 'Northern Minnesota', 'country': 'US'},
    219: {'state': 'IN', 'region': 'Northwest Indiana', 'country': 'US'},
    220: {'state': 'OH', 'region': 'Central/Southeast Ohio', 'country': 'US'},
    223: {'state': 'PA', 'region': 'South-Central Pennsylvania', 'country': 'US'},
    224: {'state': 'IL', 'region': 'Chicago suburbs', 'country': 'US'},
    225: {'state': 'LA', 'region': 'Baton Rouge', 'country': 'US'},
    226: {'state': 'ON', 'region': 'Southwest Ontario', 'country': 'CA'},
    227: {'state': 'MD', 'region': 'Western Maryland', 'country': 'US'},
    228: {'state': 'MS', 'region': 'Gulf Coast Mississippi', 'country': 'US'},
    229: {'state': 'GA', 'region': 'Southwest Georgia', 'country': 'US'},
    231: {'state': 'MI', 'region': 'Northwest Michigan', 'country': 'US'},
    234: {'state': 'OH', 'region': 'Northeast Ohio', 'country': 'US'},
    235: {'state': 'MO', 'region': 'Eastern Missouri', 'country': 'US'},
    236: {'state': 'BC', 'region': 'British Columbia', 'country': 'CA'},
    239: {'state': 'FL', 'region': 'Southwest Florida', 'country': 'US'},
    240: {'state': 'MD', 'region': 'Western Maryland', 'country': 'US'},
    242: {'state': 'BS', 'region': 'Bahamas', 'country': 'BS'},
    246: {'state': 'BB', 'region': 'Barbados', 'country': 'BB'},
    248: {'state': 'MI', 'region': 'Oakland County', 'country': 'US'},
    249: {'state': 'ON', 'region': 'Northeast Ontario', 'country': 'CA'},
    250: {'state': 'BC', 'region': 'British Columbia', 'country': 'CA'},
    251: {'state': 'AL', 'region': 'Southwest Alabama', 'country': 'US'},
    252: {'state': 'NC', 'region': 'Eastern North Carolina', 'country': 'US'},
    253: {'state': 'WA', 'region': 'Tacoma', 'country': 'US'},
    254: {'state': 'TX', 'region': 'Waco', 'country': 'US'},
    256: {'state': 'AL', 'region': 'North/East Alabama', 'country': 'US'},
    257: {'state': 'BC', 'region': 'British Columbia', 'country': 'CA'},
    260: {'state': 'IN', 'region': 'Northeast Indiana', 'country': 'US'},
    262: {'state': 'WI', 'region': 'Southeast Wisconsin', 'country': 'US'},
    263: {'state': 'QC', 'region': 'Montreal metro', 'country': 'CA'},
    264: {'state': 'AI', 'region': 'Anguilla', 'country': 'AI'},
    267: {'state': 'PA', 'region': 'Philadelphia', 'country': 'US'},
    268: {'state': 'AG', 'region': 'Antigua and Barbuda', 'country': 'AG'},
    269: {'state': 'MI', 'region': 'Southwest Michigan', 'country': 'US'},
    270: {'state': 'KY', 'region': 'Western Kentucky', 'country': 'US'},
    272: {'state': 'PA', 'region': 'Northeast Pennsylvania', 'country': 'US'},
    274: {'state': 'WI', 'region': 'Eastern Wisconsin', 'country': 'US'},
    276: {'state': 'VA', 'region': 'Southwest Virginia', 'country': 'US'},
    279: {'state': 'CA', 'region': 'Sacramento', 'country': 'US'},
    281: {'state': 'TX', 'region': 'Houston', 'country': 'US'},
    283: {'state': 'OH', 'region': 'Cincinnati', 'country': 'US'},
    284: {'state': 'VG', 'region': 'British Virgin Islands', 'country': 'VG'},
    289: {'state': 'ON', 'region': 'Golden Horseshoe', 'country': 'CA'},

    # 300s
    301: {'state': 'MD', 'region': 'Western Maryland', 'country': 'US'},
    302: {'state': 'DE', 'region': 'Delaware', 'country': 'US'},
    303: {'state': 'CO', 'region': 'Denver', 'country': 'US'},
    304: {'state': 'WV', 'region': 'West Virginia', 'country': 'US'},
    305: {'state': 'FL', 'region': 'Miami', 'country': 'US'},
    306: {'state': 'SK', 'region': 'Saskatchewan', 'country': 'CA'},
    307: {'state': 'WY', 'region': 'Wyoming', 'country': 'US'},
    308: {'state': 'NE', 'region': 'Western Nebraska', 'country': 'US'},
    309: {'state': 'IL', 'region': 'West-Central Illinois', 'country': 'US'},
    310: {'state': 'CA', 'region': 'West Los Angeles', 'country': 'US'},
    312: {'state': 'IL', 'region': 'Downtown Chicago', 'country': 'US'},
    313: {'state': 'MI', 'region': 'Detroit', 'country': 'US'},
    314: {'state': 'MO', 'region': 'St. Louis', 'country': 'US'},
    315: {'state': 'NY', 'region': 'Syracuse', 'country': 'US'},
    316: {'state': 'KS', 'region': 'Wichita', 'country': 'US'},
    317: {'state': 'IN', 'region': 'Indianapolis', 'country': 'US'},
    318: {'state': 'LA', 'region': 'Northern Louisiana', 'country': 'US'},
    319: {'state': 'IA', 'region': 'Eastern Iowa', 'country': 'US'},
    320: {'state': 'MN', 'region': 'Central Minnesota', 'country': 'US'},
    321: {'state': 'FL', 'region': 'Orlando', 'country': 'US'},
    323: {'state': 'CA', 'region': 'Los Angeles', 'country': 'US'},
    324: {'state': 'FL', 'region': 'Jacksonville', 'country': 'US'},
    325: {'state': 'TX', 'region': 'Abilene', 'country': 'US'},
    326: {'state': 'OH', 'region': 'Dayton', 'country': 'US'},
    327: {'state': 'AR', 'region': 'Central/South Arkansas', 'country': 'US'},
    329: {'state': 'NY', 'region': 'Hudson Valley', 'country': 'US'},
    330: {'state': 'OH', 'region': 'Akron/Youngstown', 'country': 'US'},
    331: {'state': 'IL', 'region': 'Chicago suburbs', 'country': 'US'},
    332: {'state': 'NY', 'region': 'Manhattan', 'country': 'US'},
    334: {'state': 'AL', 'region': 'Southeast Alabama', 'country': 'US'},
    336: {'state': 'NC', 'region': 'Piedmont Triad', 'country': 'US'},
    337: {'state': 'LA', 'region': 'Southwest Louisiana', 'country': 'US'},
    339: {'state': 'MA', 'region': 'East-Central Massachusetts', 'country': 'US'},
    340: {'state': 'VI', 'region': 'U.S. Virgin Islands', 'country': 'VI'},
    341: {'state': 'CA', 'region': 'East Bay', 'country': 'US'},
    343: {'state': 'ON', 'region': 'Ottawa', 'country': 'CA'},
    345: {'state': 'KY', 'region': 'Cayman Islands', 'country': 'KY'},
    346: {'state': 'TX', 'region': 'Houston', 'country': 'US'},
    347: {'state': 'NY', 'region': 'NYC (Bronx, Brooklyn, Queens, Staten Island)', 'country': 'US'},
    350: {'state': 'CA', 'region': 'Central California', 'country': 'US'},
    351: {'state': 'MA', 'region': 'Northeast Massachusetts', 'country': 'US'},
    352: {'state': 'FL', 'region': 'Central Florida', 'country': 'US'},
    353: {'state': 'WI', 'region': 'Southwest Wisconsin', 'country': 'US'},
    354: {'state': 'QC', 'region': 'Montreal metro', 'country': 'CA'},
    357: {'state': 'CA', 'region': 'Fresno', 'country': 'US'},
    360: {'state': 'WA', 'region': 'Western Washington', 'country': 'US'},
    361: {'state': 'TX', 'region': 'Corpus Christi', 'country': 'US'},
    363: {'state': 'NY', 'region': 'Nassau County', 'country': 'US'},
    364: {'state': 'KY', 'region': 'Western Kentucky', 'country': 'US'},
    365: {'state': 'ON', 'region': 'Golden Horseshoe', 'country': 'CA'},
    367: {'state': 'QC', 'region': 'Quebec City', 'country': 'CA'},
    368: {'state': 'AB', 'region': 'Alberta', 'country': 'CA'},
    369: {'state': 'CA', 'region': 'Northwest California', 'country': 'US'},

    # 400s
    401: {'state': 'RI', 'region': 'Rhode Island', 'country': 'US'},
    402: {'state': 'NE', 'region': 'Eastern Nebraska', 'country': 'US'},
    403: {'state': 'AB', 'region': 'Southern Alberta', 'country': 'CA'},
    404: {'state': 'GA', 'region': 'Atlanta', 'country': 'US'},
    405: {'state': 'OK', 'region': 'Central Oklahoma', 'country': 'US'},
    406: {'state': 'MT', 'region': 'Montana', 'country': 'US'},
    407: {'state': 'FL', 'region': 'Orlando', 'country': 'US'},
    408: {'state': 'CA', 'region': 'San Jose', 'country': 'US'},
    409: {'state': 'TX', 'region': 'Southeast Texas', 'country': 'US'},
    410: {'state': 'MD', 'region': 'Baltimore', 'country': 'US'},
    412: {'state': 'PA', 'region': 'Pittsburgh', 'country': 'US'},
    413: {'state': 'MA', 'region': 'Western Massachusetts', 'country': 'US'},
    414: {'state': 'WI', 'region': 'Milwaukee', 'country': 'US'},
    415: {'state': 'CA', 'region': 'San Francisco', 'country': 'US'},
    416: {'state': 'ON', 'region': 'Toronto', 'country': 'CA'},
    417: {'state': 'MO', 'region': 'Southwest Missouri', 'country': 'US'},
    418: {'state': 'QC', 'region': 'Quebec City', 'country': 'CA'},
    419: {'state': 'OH', 'region': 'Toledo', 'country': 'US'},
    423: {'state': 'TN', 'region': 'East Tennessee', 'country': 'US'},
    424: {'state': 'CA', 'region': 'West Los Angeles', 'country': 'US'},
    425: {'state': 'WA', 'region': 'Seattle suburbs', 'country': 'US'},
    428: {'state': 'NB', 'region': 'New Brunswick', 'country': 'CA'},
    430: {'state': 'TX', 'region': 'Northeast Texas', 'country': 'US'},
    431: {'state': 'MB', 'region': 'Manitoba', 'country': 'CA'},
    432: {'state': 'TX', 'region': 'West Texas', 'country': 'US'},
    434: {'state': 'VA', 'region': 'South-Central Virginia', 'country': 'US'},
    435: {'state': 'UT', 'region': 'Utah (outside Salt Lake City)', 'country': 'US'},
    436: {'state': 'OH', 'region': 'Cleveland', 'country': 'US'},
    437: {'state': 'ON', 'region': 'Toronto', 'country': 'CA'},
    438: {'state': 'QC', 'region': 'Montreal', 'country': 'CA'},
    440: {'state': 'OH', 'region': 'Cleveland suburbs', 'country': 'US'},
    441: {'state': 'BM', 'region': 'Bermuda', 'country': 'BM'},
    442: {'state': 'CA', 'region': 'Southeast California', 'country': 'US'},
    443: {'state': 'MD', 'region': 'Baltimore', 'country': 'US'},
    445: {'state': 'PA', 'region': 'Philadelphia', 'country': 'US'},
    447: {'state': 'IL', 'region': 'Central Illinois', 'country': 'US'},
    448: {'state': 'FL', 'region': 'Florida Panhandle', 'country': 'US'},
    450: {'state': 'QC', 'region': 'South of Montreal', 'country': 'CA'},
    457: {'state': 'LA', 'region': 'Northern Louisiana', 'country': 'US'},
    458: {'state': 'OR', 'region': 'Oregon (non-Portland)', 'country': 'US'},
    463: {'state': 'IN', 'region': 'Indianapolis', 'country': 'US'},
    464: {'state': 'IL', 'region': 'Chicago suburbs', 'country': 'US'},
    468: {'state': 'QC', 'region': 'Western Quebec', 'country': 'CA'},
    469: {'state': 'TX', 'region': 'Dallas', 'country': 'US'},
    470: {'state': 'GA', 'region': 'Metro Atlanta', 'country': 'US'},
    472: {'state': 'NC', 'region': 'Southeast North Carolina', 'country': 'US'},
    473: {'state': 'GD', 'region': 'Grenada', 'country': 'GD'},
    474: {'state': 'SK', 'region': 'Saskatchewan', 'country': 'CA'},
    475: {'state': 'CT', 'region': 'Southwest Connecticut', 'country': 'US'},
    478: {'state': 'GA', 'region': 'Central Georgia', 'country': 'US'},
    479: {'state': 'AR', 'region': 'Northwest Arkansas', 'country': 'US'},
    480: {'state': 'AZ', 'region': 'Phoenix metro', 'country': 'US'},
    484: {'state': 'PA', 'region': 'Southeast Pennsylvania', 'country': 'US'},

    # 500s
    501: {'state': 'AR', 'region': 'Central Arkansas', 'country': 'US'},
    502: {'state': 'KY', 'region': 'Louisville', 'country': 'US'},
    503: {'state': 'OR', 'region': 'Portland', 'country': 'US'},
    504: {'state': 'LA', 'region': 'New Orleans', 'country': 'US'},
    505: {'state': 'NM', 'region': 'New Mexico', 'country': 'US'},
    506: {'state': 'NB', 'region': 'New Brunswick', 'country': 'CA'},
    507: {'state': 'MN', 'region': 'Southern Minnesota', 'country': 'US'},
    508: {'state': 'MA', 'region': 'Southeast Massachusetts', 'country': 'US'},
    509: {'state': 'WA', 'region': 'Eastern Washington', 'country': 'US'},
    510: {'state': 'CA', 'region': 'East Bay', 'country': 'US'},
    512: {'state': 'TX', 'region': 'Austin', 'country': 'US'},
    513: {'state': 'OH', 'region': 'Cincinnati', 'country': 'US'},
    514: {'state': 'QC', 'region': 'Montreal', 'country': 'CA'},
    515: {'state': 'IA', 'region': 'Central Iowa', 'country': 'US'},
    516: {'state': 'NY', 'region': 'Nassau County', 'country': 'US'},
    517: {'state': 'MI', 'region': 'South-Central Michigan', 'country': 'US'},
    518: {'state': 'NY', 'region': 'Northeast New York', 'country': 'US'},
    519: {'state': 'ON', 'region': 'Southwest Ontario', 'country': 'CA'},

    # 600s
    601: {'state': 'MS', 'region': 'Central Mississippi', 'country': 'US'},
    602: {'state': 'AZ', 'region': 'Phoenix', 'country': 'US'},
    603: {'state': 'NH', 'region': 'New Hampshire', 'country': 'US'},
    604: {'state': 'BC', 'region': 'Lower Mainland', 'country': 'CA'},
    605: {'state': 'SD', 'region': 'South Dakota', 'country': 'US'},
    606: {'state': 'KY', 'region': 'Eastern Kentucky', 'country': 'US'},
    607: {'state': 'NY', 'region': 'Southern New York', 'country': 'US'},
    608: {'state': 'WI', 'region': 'Southwest Wisconsin', 'country': 'US'},
    609: {'state': 'NJ', 'region': 'South Jersey', 'country': 'US'},
    610: {'state': 'PA', 'region': 'Southeast Pennsylvania', 'country': 'US'},
    612: {'state': 'MN', 'region': 'Minneapolis', 'country': 'US'},
    613: {'state': 'ON', 'region': 'Ottawa', 'country': 'CA'},
    614: {'state': 'OH', 'region': 'Columbus', 'country': 'US'},
    615: {'state': 'TN', 'region': 'Nashville', 'country': 'US'},
    616: {'state': 'MI', 'region': 'Grand Rapids', 'country': 'US'},
    617: {'state': 'MA', 'region': 'Boston', 'country': 'US'},
    618: {'state': 'IL', 'region': 'Southern Illinois', 'country': 'US'},
    619: {'state': 'CA', 'region': 'San Diego', 'country': 'US'},
    620: {'state': 'KS', 'region': 'Southern Kansas', 'country': 'US'},
    621: {'state': 'NC', 'region': 'North Carolina', 'country': 'US'},
    623: {'state': 'AZ', 'region': 'Phoenix suburbs', 'country': 'US'},
    626: {'state': 'CA', 'region': 'San Gabriel Valley', 'country': 'US'},
    628: {'state': 'CA', 'region': 'San Francisco', 'country': 'US'},
    629: {'state': 'TN', 'region': 'Nashville', 'country': 'US'},
    630: {'state': 'IL', 'region': 'Chicago western suburbs', 'country': 'US'},
    631: {'state': 'NY', 'region': 'Suffolk County', 'country': 'US'},
    636: {'state': 'MO', 'region': 'St. Louis suburbs', 'country': 'US'},
    639: {'state': 'SK', 'region': 'Saskatchewan', 'country': 'CA'},
    641: {'state': 'IA', 'region': 'Central Iowa', 'country': 'US'},
    645: {'state': 'NY', 'region': 'Manhattan', 'country': 'US'},
    647: {'state': 'ON', 'region': 'Toronto', 'country': 'CA'},
    650: {'state': 'CA', 'region': 'San Mateo County', 'country': 'US'},
    659: {'state': 'AL', 'region': 'North Alabama', 'country': 'US'},
    661: {'state': 'CA', 'region': 'Bakersfield', 'country': 'US'},
    662: {'state': 'MS', 'region': 'North Mississippi', 'country': 'US'},
    667: {'state': 'MD', 'region': 'Baltimore', 'country': 'US'},
    669: {'state': 'CA', 'region': 'San Jose', 'country': 'US'},
    670: {'state': 'MP', 'region': 'Northern Mariana Islands', 'country': 'MP'},
    671: {'state': 'GU', 'region': 'Guam', 'country': 'GU'},
    672: {'state': 'BC', 'region': 'British Columbia', 'country': 'CA'},
    673: {'state': 'BC', 'region': 'British Columbia', 'country': 'CA'},
    678: {'state': 'GA', 'region': 'Metro Atlanta', 'country': 'US'},
    679: {'state': 'MI', 'region': 'Detroit', 'country': 'US'},
    680: {'state': 'NY', 'region': 'Syracuse', 'country': 'US'},
    681: {'state': 'WV', 'region': 'West Virginia', 'country': 'US'},
    682: {'state': 'TX', 'region': 'Fort Worth', 'country': 'US'},
    683: {'state': 'ON', 'region': 'Ottawa', 'country': 'CA'},
    684: {'state': 'AS', 'region': 'American Samoa', 'country': 'AS'},
    689: {'state': 'FL', 'region': 'Central Florida', 'country': 'US'},

    # 700s
    701: {'state': 'ND', 'region': 'North Dakota', 'country': 'US'},
    702: {'state': 'NV', 'region': 'Las Vegas', 'country': 'US'},
    703: {'state': 'VA', 'region': 'Northern Virginia', 'country': 'US'},
    704: {'state': 'NC', 'region': 'Charlotte', 'country': 'US'},
    705: {'state': 'ON', 'region': 'Northern Ontario', 'country': 'CA'},
    706: {'state': 'GA', 'region': 'North Georgia', 'country': 'US'},
    707: {'state': 'CA', 'region': 'Northwest California', 'country': 'US'},
    708: {'state': 'IL', 'region': 'Chicago suburbs', 'country': 'US'},
    709: {'state': 'NL', 'region': 'Newfoundland and Labrador', 'country': 'CA'},
    712: {'state': 'IA', 'region': 'Western Iowa', 'country': 'US'},
    713: {'state': 'TX', 'region': 'Houston', 'country': 'US'},
    714: {'state': 'CA', 'region': 'Orange County', 'country': 'US'},
    715: {'state': 'WI', 'region': 'Northern Wisconsin', 'country': 'US'},
    716: {'state': 'NY', 'region': 'Buffalo', 'country': 'US'},
    717: {'state': 'PA', 'region': 'South-Central Pennsylvania', 'country': 'US'},
    718: {'state': 'NY', 'region': 'NYC (Bronx, Brooklyn, Queens, Staten Island)', 'country': 'US'},
    719: {'state': 'CO', 'region': 'Colorado Springs', 'country': 'US'},
    720: {'state': 'CO', 'region': 'Denver', 'country': 'US'},
    721: {'state': 'SX', 'region': 'Sint Maarten', 'country': 'SX'},
    724: {'state': 'PA', 'region': 'Southwest Pennsylvania', 'country': 'US'},
    725: {'state': 'NV', 'region': 'Las Vegas', 'country': 'US'},
    726: {'state': 'TX', 'region': 'San Antonio', 'country': 'US'},
    727: {'state': 'FL', 'region': 'St. Petersburg', 'country': 'US'},
    728: {'state': 'FL', 'region': 'Central Florida', 'country': 'US'},
    729: {'state': 'MS', 'region': 'North Mississippi', 'country': 'US'},
    731: {'state': 'TN', 'region': 'West Tennessee', 'country': 'US'},
    732: {'state': 'NJ', 'region': 'Central New Jersey', 'country': 'US'},
    734: {'state': 'MI', 'region': 'Southeast Michigan', 'country': 'US'},
    737: {'state': 'TX', 'region': 'Austin', 'country': 'US'},
    738: {'state': 'AB', 'region': 'Alberta', 'country': 'CA'},
    740: {'state': 'OH', 'region': 'Southeast Ohio', 'country': 'US'},
    742: {'state': 'ON', 'region': 'Southwest Ontario', 'country': 'CA'},
    743: {'state': 'NC', 'region': 'Piedmont Triad', 'country': 'US'},
    747: {'state': 'CA', 'region': 'San Fernando Valley', 'country': 'US'},
    751: {'state': 'RI', 'region': 'Rhode Island', 'country': 'US'},
    752: {'state': 'CA', 'region': 'Central California', 'country': 'US'},
    753: {'state': 'ON', 'region': 'Ottawa', 'country': 'CA'},
    754: {'state': 'FL', 'region': 'Broward County', 'country': 'US'},
    757: {'state': 'VA', 'region': 'Tidewater Virginia', 'country': 'US'},
    760: {'state': 'CA', 'region': 'North San Diego County', 'country': 'US'},
    762: {'state': 'GA', 'region': 'North Georgia', 'country': 'US'},
    763: {'state': 'MN', 'region': 'Minneapolis suburbs', 'country': 'US'},
    765: {'state': 'IN', 'region': 'Central Indiana', 'country': 'US'},
    770: {'state': 'GA', 'region': 'Atlanta suburbs', 'country': 'US'},
    771: {'state': 'DC', 'region': 'Washington, D.C.', 'country': 'US'},
    772: {'state': 'FL', 'region': 'Treasure Coast', 'country': 'US'},
    773: {'state': 'IL', 'region': 'Chicago', 'country': 'US'},
    774: {'state': 'MA', 'region': 'Southeast Massachusetts', 'country': 'US'},
    775: {'state': 'NV', 'region': 'Northern Nevada', 'country': 'US'},
    778: {'state': 'BC', 'region': 'British Columbia', 'country': 'CA'},
    779: {'state': 'IL', 'region': 'Northern Illinois', 'country': 'US'},
    781: {'state': 'MA', 'region': 'Greater Boston', 'country': 'US'},
    782: {'state': 'NS', 'region': 'Nova Scotia', 'country': 'CA'},
    785: {'state': 'KS', 'region': 'Northern Kansas', 'country': 'US'},
    786: {'state': 'FL', 'region': 'Miami', 'country': 'US'},

    # 800s
    801: {'state': 'UT', 'region': 'Salt Lake City', 'country': 'US'},
    802: {'state': 'VT', 'region': 'Vermont', 'country': 'US'},
    803: {'state': 'SC', 'region': 'Central South Carolina', 'country': 'US'},
    804: {'state': 'VA', 'region': 'Central Virginia', 'country': 'US'},
    805: {'state': 'CA', 'region': 'Central California Coast', 'country': 'US'},
    806: {'state': 'TX', 'region': 'Panhandle Texas', 'country': 'US'},
    807: {'state': 'ON', 'region': 'Northwest Ontario', 'country': 'CA'},
    808: {'state': 'HI', 'region': 'Hawaii', 'country': 'US'},
    810: {'state': 'MI', 'region': 'Southeast Michigan', 'country': 'US'},
    812: {'state': 'IN', 'region': 'Southern Indiana', 'country': 'US'},
    813: {'state': 'FL', 'region': 'Tampa', 'country': 'US'},
    814: {'state': 'PA', 'region': 'Northwest Pennsylvania', 'country': 'US'},
    815: {'state': 'IL', 'region': 'Northern Illinois', 'country': 'US'},
    816: {'state': 'MO', 'region': 'Kansas City', 'country': 'US'},
    817: {'state': 'TX', 'region': 'Fort Worth', 'country': 'US'},
    818: {'state': 'CA', 'region': 'San Fernando Valley', 'country': 'US'},
    819: {'state': 'QC', 'region': 'Western Quebec', 'country': 'CA'},
    825: {'state': 'AB', 'region': 'Alberta', 'country': 'CA'},
    828: {'state': 'NC', 'region': 'Western North Carolina', 'country': 'US'},
    830: {'state': 'TX', 'region': 'Southwest Texas', 'country': 'US'},
    831: {'state': 'CA', 'region': 'Monterey Bay', 'country': 'US'},
    832: {'state': 'TX', 'region': 'Houston', 'country': 'US'},
    835: {'state': 'PA', 'region': 'Southeast Pennsylvania', 'country': 'US'},
    838: {'state': 'NY', 'region': 'Hudson Valley', 'country': 'US'},
    839: {'state': 'SC', 'region': 'Central South Carolina', 'country': 'US'},
    840: {'state': 'CA', 'region': 'Sacramento', 'country': 'US'},
    843: {'state': 'SC', 'region': 'Coastal South Carolina', 'country': 'US'},
    845: {'state': 'NY', 'region': 'Hudson Valley', 'country': 'US'},
    847: {'state': 'IL', 'region': 'Chicago northern suburbs', 'country': 'US'},
    848: {'state': 'NJ', 'region': 'Central New Jersey', 'country': 'US'},
    850: {'state': 'FL', 'region': 'Florida Panhandle', 'country': 'US'},
    854: {'state': 'SC', 'region': 'Coastal South Carolina', 'country': 'US'},
    856: {'state': 'NJ', 'region': 'South Jersey', 'country': 'US'},
    857: {'state': 'MA', 'region': 'Boston', 'country': 'US'},
    858: {'state': 'CA', 'region': 'San Diego', 'country': 'US'},
    859: {'state': 'KY', 'region': 'Lexington', 'country': 'US'},
    860: {'state': 'CT', 'region': 'Connecticut', 'country': 'US'},
    862: {'state': 'NJ', 'region': 'North Jersey', 'country': 'US'},
    863: {'state': 'FL', 'region': 'Central Florida', 'country': 'US'},
    864: {'state': 'SC', 'region': 'Upstate South Carolina', 'country': 'US'},
    865: {'state': 'TN', 'region': 'Knoxville', 'country': 'US'},
    867: {'state': 'YT/NT/NU', 'region': 'Yukon/Northwest Territories/Nunavut', 'country': 'CA'},
    870: {'state': 'AR', 'region': 'Eastern Arkansas', 'country': 'US'},
    872: {'state': 'IL', 'region': 'Downtown Chicago', 'country': 'US'},
    873: {'state': 'QC', 'region': 'Quebec City', 'country': 'CA'},
    878: {'state': 'PA', 'region': 'Pittsburgh', 'country': 'US'},

    # 900s
    901: {'state': 'TN', 'region': 'Memphis', 'country': 'US'},
    902: {'state': 'NS/PE', 'region': 'Nova Scotia/Prince Edward Island', 'country': 'CA'},
    903: {'state': 'TX', 'region': 'Northeast Texas', 'country': 'US'},
    904: {'state': 'FL', 'region': 'Jacksonville', 'country': 'US'},
    905: {'state': 'ON', 'region': 'Golden Horseshoe', 'country': 'CA'},
    906: {'state': 'MI', 'region': 'Upper Peninsula', 'country': 'US'},
    907: {'state': 'AK', 'region': 'Alaska', 'country': 'US'},
    908: {'state': 'NJ', 'region': 'Central New Jersey', 'country': 'US'},
    909: {'state': 'CA', 'region': 'Inland Empire', 'country': 'US'},
    910: {'state': 'NC', 'region': 'Southeast North Carolina', 'country': 'US'},
    912: {'state': 'GA', 'region': 'Southeast Georgia', 'country': 'US'},
    913: {'state': 'KS', 'region': 'Kansas City metro', 'country': 'US'},
    914: {'state': 'NY', 'region': 'Westchester County', 'country': 'US'},
    915: {'state': 'TX', 'region': 'El Paso', 'country': 'US'},
    916: {'state': 'CA', 'region': 'Sacramento', 'country': 'US'},
    917: {'state': 'NY', 'region': 'New York City', 'country': 'US'},
    918: {'state': 'OK', 'region': 'Tulsa', 'country': 'US'},
    919: {'state': 'NC', 'region': 'Research Triangle', 'country': 'US'},
    920: {'state': 'WI', 'region': 'Northeast Wisconsin', 'country': 'US'},
    925: {'state': 'CA', 'region': 'Contra Costa County', 'country': 'US'},
    928: {'state': 'AZ', 'region': 'Northern Arizona', 'country': 'US'},
    929: {'state': 'NY', 'region': 'Queens/Bronx', 'country': 'US'},
    930: {'state': 'IN', 'region': 'Southwest Indiana', 'country': 'US'},
    931: {'state': 'TN', 'region': 'Middle Tennessee', 'country': 'US'},
    934: {'state': 'NY', 'region': 'Nassau/Suffolk Counties', 'country': 'US'},
    936: {'state': 'TX', 'region': 'East Texas', 'country': 'US'},
    937: {'state': 'OH', 'region': 'Southwest Ohio', 'country': 'US'},
    938: {'state': 'AL', 'region': 'North Alabama', 'country': 'US'},
    940: {'state': 'TX', 'region': 'North Texas', 'country': 'US'},
    941: {'state': 'FL', 'region': 'Southwest Florida', 'country': 'US'},
    942: {'state': 'AB', 'region': 'Alberta', 'country': 'CA'},
    945: {'state': 'TX', 'region': 'Dallas metro', 'country': 'US'},
    947: {'state': 'MI', 'region': 'Detroit metro', 'country': 'US'},
    949: {'state': 'CA', 'region': 'Orange County', 'country': 'US'},
    951: {'state': 'CA', 'region': 'Riverside/San Bernardino', 'country': 'US'},
    952: {'state': 'MN', 'region': 'Minneapolis southern suburbs', 'country': 'US'},
    954: {'state': 'FL', 'region': 'Broward County', 'country': 'US'},
    956: {'state': 'TX', 'region': 'South Texas', 'country': 'US'},
    959: {'state': 'CT', 'region': 'Connecticut', 'country': 'US'},
    970: {'state': 'CO', 'region': 'Northern Colorado', 'country': 'US'},
    971: {'state': 'OR', 'region': 'Portland', 'country': 'US'},
    972: {'state': 'TX', 'region': 'Dallas', 'country': 'US'},
    973: {'state': 'NJ', 'region': 'North Jersey', 'country': 'US'},
    978: {'state': 'MA', 'region': 'Northeast Massachusetts', 'country': 'US'},
    979: {'state': 'TX', 'region': 'Bryan/College Station', 'country': 'US'},
    980: {'state': 'NC', 'region': 'Charlotte', 'country': 'US'},
    983: {'state': 'CO', 'region': 'Denver', 'country': 'US'},
    984: {'state': 'NC', 'region': 'Research Triangle', 'country': 'US'},
    985: {'state': 'LA', 'region': 'Southeast Louisiana', 'country': 'US'},
    986: {'state': 'ID', 'region': 'Idaho', 'country': 'US'},
    989: {'state': 'MI', 'region': 'Central Michigan', 'country': 'US'},
}


def get_location_info(area_code):
    """
    Get geographic location information for an area code.

    Args:
        area_code: 3-digit area code (int or str)

    Returns:
        dict or None: {
            'state': str (2-letter state/province code),
            'region': str (descriptive region name),
            'country': str (2-letter country code: US, CA, etc.)
        }
    """
    try:
        code = int(area_code)
        return AREA_CODE_MAPPING.get(code)
    except (ValueError, TypeError):
        return None



from typing import Optional, List, Set
from contextlib import contextmanager
from django.db import models
import redis.exceptions
from .client import get_connection
from ...errors import TimeoutException


class RedisBasePool:
    """Simple Redis pool using atomic Redis operations."""

    def __init__(self, pool_key: str, default_timeout: int = 30):
        """
        Initialize the Redis pool.

        Args:
            pool_key: Unique identifier for this pool
            default_timeout: Default timeout in seconds for blocking operations
        """
        self.pool_key = pool_key
        self.default_timeout = default_timeout
        self.redis_client = get_connection()

        self.available_list_key = f"{pool_key}:list"
        self.all_items_set_key = f"{pool_key}:set"

    def is_ready(self) -> bool:
        """Check if the pool is ready."""
        return self.redis_client.exists(self.available_list_key) and self.redis_client.exists(self.all_items_set_key)

    def add(self, str_id: str) -> bool:
        """Add an item to the pool."""
        if not isinstance(str_id, str):
            str_id = str(str_id)
        if not self.redis_client.sismember(self.all_items_set_key, str_id):
            self.redis_client.sadd(self.all_items_set_key, str_id)
            self.redis_client.lpush(self.available_list_key, str_id)
            return True
        return False

    def remove(self, str_id: str, force: bool = False) -> bool:
        """
        Remove an item from the pool entirely.

        Args:
            str_id: Item to remove
            force: If True, removes even if checked out. If False (default),
                   only removes if currently available.

        Returns:
            True if removed, False if not in pool or checked out (when force=False)
        """
        if not isinstance(str_id, str):
            str_id = str(str_id)
        if not self.redis_client.sismember(self.all_items_set_key, str_id):
            return False

        if not force:
            # Check if item is available (not checked out)
            available = self.redis_client.lrange(self.available_list_key, 0, -1)
            if str_id.encode() not in available and str_id not in available:
                return False

        self.redis_client.srem(self.all_items_set_key, str_id)
        self.redis_client.lrem(self.available_list_key, 0, str_id)
        return True

    def clear(self) -> None:
        """Clear all items from the pool."""
        self.redis_client.delete(self.available_list_key)
        self.redis_client.delete(self.all_items_set_key)

    def checkout(self, str_id: str, timeout: Optional[int] = None) -> bool:
        """Check out a specific item from the pool."""
        if not isinstance(str_id, str):
            str_id = str(str_id)
        if not self.redis_client.sismember(self.all_items_set_key, str_id):
            return False

        if timeout is None:
            removed = self.redis_client.lrem(self.available_list_key, 1, str_id)
            return removed > 0

        import time
        start = time.time()
        while self.redis_client.lrem(self.available_list_key, 1, str_id) == 0:
            time.sleep(1.0)
            elapsed = time.time() - start
            if elapsed > timeout:
                return False
        return True

    def checkin(self, str_id: str, allow_duplicate: bool = False) -> bool:
        """
        Check in an item back to the pool.

        Args:
            str_id: Item to check in
            allow_duplicate: If False (default), checks if item is already available
                           to prevent duplicates. If True, always adds to list.

        Returns:
            True if item was checked in, False if item not in pool or already available
        """
        if not isinstance(str_id, str):
            str_id = str(str_id)
        if not self.redis_client.sismember(self.all_items_set_key, str_id):
            return False

        # Prevent duplicates by checking if item is already in available list
        if not allow_duplicate:
            # Check if item is in the available list
            available = self.redis_client.lrange(self.available_list_key, 0, -1)
            if str_id.encode() in available or str_id in available:
                return False

        self.redis_client.lpush(self.available_list_key, str_id)
        return True

    def list_all(self) -> Set[str]:
        """List all items in the pool."""
        return self.redis_client.smembers(self.all_items_set_key)

    def list_available(self) -> List[str]:
        """List available items in the pool."""
        return self.redis_client.lrange(self.available_list_key, 0, -1)

    def list_checked_out(self) -> Set[str]:
        """List checked out items."""
        all_items = self.list_all()
        available_items = set(self.list_available())
        return all_items - available_items

    def destroy_pool(self) -> None:
        """Completely destroy the pool."""
        self.clear()

    def remove_duplicates(self) -> int:
        """
        Remove duplicate entries from the available list.

        Returns:
            Number of duplicates removed
        """
        available = self.redis_client.lrange(self.available_list_key, 0, -1)
        seen = set()
        duplicates_removed = 0

        # Clear the list
        self.redis_client.delete(self.available_list_key)

        # Re-add unique items only
        for item in reversed(available):  # Reversed to maintain original order
            if item not in seen:
                seen.add(item)
                self.redis_client.lpush(self.available_list_key, item)
            else:
                duplicates_removed += 1

        return duplicates_removed

    def get_next_available(self, timeout: Optional[int] = None) -> Optional[str]:
        """Get the next available item from the pool."""
        timeout = timeout or self.default_timeout

        try:
            result = self.redis_client.brpop(self.available_list_key, timeout=timeout)
            if result:
                return result[1]
            return None
        except redis.exceptions.TimeoutError:
            return None

    @contextmanager
    def checkout_item(self, timeout: Optional[int] = None, raise_on_timeout: bool = True):
        """Context manager for safely checking out and returning items."""
        item = self.get_next_available(timeout)
        if item is None:
            if raise_on_timeout:
                raise RuntimeError("No items available in pool")
            else:
                yield None
                return

        try:
            yield item
        finally:
            self.checkin(item)

    @contextmanager
    def checkout_specific_item(self, str_id: str, timeout: Optional[int] = None):
        """Context manager for checking out a specific item."""
        if not isinstance(str_id, str):
            str_id = str(str_id)
        if not self.checkout(str_id, timeout):
            raise RuntimeError(f"Could not checkout item {str_id}")

        try:
            yield str_id
        finally:
            self.checkin(str_id)


class RedisModelPool(RedisBasePool):
    """Django model-specific Redis pool."""

    def __init__(self, model_cls, query_dict, pool_key=None, default_timeout=30):
        """
        Initialize the model pool.

        Args:
            model_cls: Django model class
            query_dict: Query parameters to filter model instances
            pool_key: Unique identifier for this pool
            default_timeout: Default timeout in seconds
        """
        if pool_key is None:
            pool_key = f"modelpool:{model_cls.__name__}"
        super().__init__(pool_key, default_timeout)
        self.model_cls = model_cls
        self.query_dict = query_dict

    def init_pool(self) -> None:
        """Initialize pool with model instances."""
        self.destroy_pool()

        queryset = self.model_cls.objects.filter(**self.query_dict)
        for instance in queryset:
            item = str(instance.pk)
            # After destroy_pool(), the set is empty, so no need to check membership
            self.redis_client.sadd(self.all_items_set_key, item)
            self.redis_client.lpush(self.available_list_key, item)

    def add_to_pool(self, instance: models.Model) -> bool:
        """
        Add a model instance to the pool.

        Returns:
            True if item was added (either by init_pool or directly), False if it already existed
        """
        item = str(instance.pk)

        # Check if item exists before potential init
        existed_before = self.redis_client.exists(self.all_items_set_key) and \
                        self.redis_client.sismember(self.all_items_set_key, item)

        if not self.is_ready():
            self.init_pool()
            # If item didn't exist before but exists now, init_pool added it
            if not existed_before and self.redis_client.sismember(self.all_items_set_key, item):
                return True

        if not self.redis_client.sismember(self.all_items_set_key, item):
            self.redis_client.sadd(self.all_items_set_key, item)
            self.redis_client.lpush(self.available_list_key, item)
            return True
        return False

    def remove_from_pool(self, instance: models.Model, force: bool = False) -> bool:
        """
        Remove instance from pool.

        Args:
            instance: Model instance to remove
            force: If True, removes even if checked out. If False (default),
                   only removes if currently available.

        Returns:
            True if removed, False if not in pool or checked out (when force=False)
        """
        if not self.redis_client.exists(self.all_items_set_key):
            self.init_pool()

        item = str(instance.pk)
        if not self.redis_client.sismember(self.all_items_set_key, item):
            return False

        if not force:
            # Check if item is available (not checked out)
            available = self.redis_client.lrange(self.available_list_key, 0, -1)
            if item.encode() not in available and item not in available:
                return False

        self.redis_client.lrem(self.available_list_key, 0, item)
        self.redis_client.srem(self.all_items_set_key, item)
        return True

    def list_checked_out_instances(self):
        """List checked out items."""
        all_items = self.list_checked_out()
        return self.model_cls.objects.filter(pk__in=all_items)

    def get_next_instance(self, timeout: Optional[int] = None, _retries: int = 0, _max_retries: int = 100) -> Optional[models.Model]:
        """
        Get the next available model instance.

        Args:
            timeout: Timeout for waiting for an available instance
            _retries: Internal counter to prevent infinite recursion
            _max_retries: Maximum number of retries for stale records
        """
        if not self.redis_client.exists(self.all_items_set_key):
            self.init_pool()

        if _retries >= _max_retries:
            return None

        pk = self.get_next_available(timeout)
        if pk:
            try:
                instance = self.model_cls.objects.get(pk=pk)

                # Verify instance still matches criteria
                for key, value in self.query_dict.items():
                    if getattr(instance, key) != value:
                        self.redis_client.srem(self.all_items_set_key, pk)
                        return self.get_next_instance(timeout=0, _retries=_retries + 1, _max_retries=_max_retries)

                return instance
            except self.model_cls.DoesNotExist:
                self.redis_client.srem(self.all_items_set_key, pk)
                return self.get_next_instance(timeout=0, _retries=_retries + 1, _max_retries=_max_retries)

        return None

    def get_specific_instance(self, instance: models.Model) -> bool:
        """Get a specific instance from pool."""
        item = str(instance.pk)
        if not self.redis_client.sismember(self.all_items_set_key, item):
            return False

        removed = self.redis_client.lrem(self.available_list_key, 1, item)
        return removed > 0

    def return_instance(self, instance: models.Model, allow_duplicate: bool = False) -> bool:
        """
        Return a model instance to the pool.

        Args:
            instance: Model instance to return
            allow_duplicate: If False (default), prevents returning an already-available instance

        Returns:
            True if returned, False if not in pool or already available
        """
        item = str(instance.pk)
        if not self.redis_client.sismember(self.all_items_set_key, item):
            return False

        # Prevent duplicates by checking if item is already in available list
        if not allow_duplicate:
            available = self.redis_client.lrange(self.available_list_key, 0, -1)
            if item.encode() in available or item in available:
                return False

        self.redis_client.lpush(self.available_list_key, item)
        return True

    @contextmanager
    def checkout_instance(self, timeout: Optional[int] = None):
        """Context manager for safely checking out and returning model instances."""
        instance = self.get_next_instance(timeout)
        if instance is None:
            raise RuntimeError("No instances available in pool")

        try:
            yield instance
        finally:
            self.return_instance(instance)

    @contextmanager
    def checkout_specific_instance(self, instance: models.Model):
        """Context manager for checking out a specific model instance."""
        if not self.get_specific_instance(instance):
            raise RuntimeError(f"Could not checkout instance {instance.pk}")

        try:
            yield instance
        finally:
            self.return_instance(instance)


# Example usage:
if __name__ == "__main__":
    # Basic pool
    pool = RedisBasePool("test_pool")

    # Add items
    pool.add("item1")
    pool.add("item2")
    pool.add("item3")

    print("All items:", pool.list_all())
    print("Available:", pool.list_available())

    # Using context manager (recommended)
    try:
        with pool.checkout_item(timeout=5) as item:
            print(f"Got item: {item}")
            # Do work with item - automatically returned even if exception occurs
    except RuntimeError as e:
        print(f"No items available: {e}")

    # Django model example:
    # model_pool = RedisModelPool(
    #     model_cls=MyModel,
    #     query_dict={"status": "active"},
    #     pool_key="active_models"
    # )
    #
    # # Initialize pool
    # model_pool.init_pool()
    #
    # # Using context manager (recommended)
    # try:
    #     with model_pool.checkout_instance(timeout=30) as instance:
    #         print(f"Got instance: {instance}")
    #         # Do work with instance - automatically returned
    # except RuntimeError as e:
    #     print(f"No instances available: {e}")



from .redis_metrics import (
    record,
    fetch,
    fetch_values,
    get_categories,
    fetch_by_category,
    get_category_slugs,
    delete_category,
    get_view_perms,
    get_write_perms,
    set_view_perms,
    set_write_perms,
    list_accounts,
    add_account,
    delete_account,
    get_accounts_with_permissions,
    set_value,
    get_value
)



import datetime
import importlib
from typing import Callable, List, Dict, Any
from django.apps import apps
from mojo.decorators.cron import schedule

def run_now() -> None:
    """
    Execute the scheduled functions that match the current time.

    Retrieves the list of functions scheduled to run at the current
    date and time, and executes each of them.
    """
    functions_to_run = find_scheduled_functions()
    for func in functions_to_run:
        func()

def find_scheduled_functions() -> List[Callable]:
    """
    Find all functions that are scheduled to run at the current time.

    Returns:
        List[Callable]: A list of callable functions that match the
        current date and time according to their cron specifications.
    """
    if not hasattr(schedule, 'scheduled_functions'):
        return []

    now = datetime.datetime.now()
    matching_funcs = []

    for cron_spec in schedule.scheduled_functions:
        if match_time(now, cron_spec):
            matching_funcs.append(cron_spec['func'])

    return matching_funcs

def match_time(current_time: datetime.datetime, cron_spec: Dict[str, Any]) -> bool:
    """
    Determine if a given time matches a cron-like specification.

    Args:
        current_time (datetime.datetime): The current date and time.
        cron_spec (Dict[str, Any]): A dictionary containing the cron
        specifications to match against.

    Returns:
        bool: True if the current time matches the cron specification,
        False otherwise.
    """
    cron_field = {
        'minutes': current_time.minute,
        'hours': current_time.hour,
        'days': current_time.day,
        'months': current_time.month,
        'weekdays': current_time.weekday()
    }

    for field, time_value in cron_field.items():
        if not matches(cron_spec[field], time_value):
            return False
    return True

def matches(cron_value: str, actual_value: int) -> bool:
    """
    Check if a specific time value matches the corresponding cron pattern.

    Args:
        cron_value (str): The cron pattern to match. Supports:
            - '*' for wildcard (matches all values)
            - '5' for specific value
            - '1,3,5' for comma-separated values
            - '1-5' for ranges (inclusive)
            - '*/5' for steps (every 5th value)
            - '10-50/5' for ranges with steps
        actual_value (int): The actual time value to compare.

    Returns:
        bool: True if the actual value matches the cron pattern,
        False otherwise.
    """
    if cron_value == '*':
        return True
    if isinstance(cron_value, str):
        patterns = cron_value.split(',')
    else:
        patterns = cron_value

    # Split by comma to handle multiple patterns
    for pattern in patterns:
        pattern = pattern.strip()

        # Handle step values (e.g., */5, 10-50/5)
        if '/' in pattern:
            range_part, step = pattern.split('/', 1)
            try:
                step_value = int(step)
            except ValueError:
                continue

            # Handle */step (wildcard with step)
            if range_part == '*':
                if actual_value % step_value == 0:
                    return True
            # Handle range/step (e.g., 10-50/5)
            elif '-' in range_part:
                start, end = range_part.split('-', 1)
                try:
                    start_val = int(start)
                    end_val = int(end)
                    if start_val <= actual_value <= end_val and \
                       (actual_value - start_val) % step_value == 0:
                        return True
                except ValueError:
                    continue

        # Handle ranges (e.g., 1-5)
        elif '-' in pattern:
            start, end = pattern.split('-', 1)
            try:
                start_val = int(start)
                end_val = int(end)
                if start_val <= actual_value <= end_val:
                    return True
            except ValueError:
                continue

        # Handle single values
        else:
            try:
                if int(pattern) == actual_value:
                    return True
            except ValueError:
                continue

    return False

def load_app_cron() -> None:
    """
    Load cronjob modules from all Django apps.

    Iterates through each registered Django app and attempts to import
    the APP_NAME.cronjobs module if it exists. This ensures that any
    cron-decorated functions are properly registered with the scheduler.
    """
    for app_config in apps.get_app_configs():
        app_name = app_config.name
        cronjobs_module = f"{app_name}.cronjobs"
        try:
            importlib.import_module(cronjobs_module)
        except ImportError:
            # App doesn't have a cronjobs module, continue to next app
            continue



# Generated by Django 5.2.12 on 2026-03-27 03:18

import django.db.models.deletion
import mojo.models.rest
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('account', '0036_bouncerdevice_userdevice_muid_botsignature_and_more'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ChatRoom',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, default='', max_length=255)),
                ('kind', models.CharField(choices=[('direct', 'Direct Message'), ('group', 'Group Chat'), ('channel', 'Channel')], db_index=True, default='group', max_length=20)),
                ('rules', models.JSONField(blank=True, default=dict)),
                ('metadata', models.JSONField(blank=True, default=dict)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True, db_index=True)),
                ('group', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='chat_rooms', to='account.group')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='owned_chat_rooms', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-modified'],
            },
            bases=(models.Model, mojo.models.rest.MojoModel),
        ),
        migrations.CreateModel(
            name='ChatMessage',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('body', models.TextField()),
                ('kind', models.CharField(choices=[('text', 'Text'), ('image', 'Image'), ('system', 'System')], default='text', max_length=20)),
                ('moderation_decision', models.CharField(choices=[('allow', 'Allow'), ('warn', 'Warn'), ('block', 'Block')], default='allow', max_length=10)),
                ('edited_at', models.DateTimeField(blank=True, null=True)),
                ('is_flagged', models.BooleanField(db_index=True, default=False)),
                ('flagged_at', models.DateTimeField(blank=True, null=True)),
                ('metadata', models.JSONField(blank=True, default=dict)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('flagged_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='chat_messages', to=settings.AUTH_USER_MODEL)),
                ('room', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='messages', to='chat.chatroom')),
            ],
            options={
                'ordering': ['-created'],
            },
            bases=(models.Model, mojo.models.rest.MojoModel),
        ),
        migrations.CreateModel(
            name='ChatMembership',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('role', models.CharField(choices=[('member', 'Member'), ('admin', 'Admin'), ('owner', 'Owner')], default='member', max_length=20)),
                ('status', models.CharField(choices=[('active', 'Active'), ('muted', 'Muted'), ('banned', 'Banned')], default='active', max_length=20)),
                ('last_read_at', models.DateTimeField(blank=True, null=True)),
                ('joined_at', models.DateTimeField(auto_now_add=True)),
                ('metadata', models.JSONField(blank=True, default=dict)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='chat_memberships', to=settings.AUTH_USER_MODEL)),
                ('room', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='memberships', to='chat.chatroom')),
            ],
            bases=(models.Model, mojo.models.rest.MojoModel),
        ),
        migrations.CreateModel(
            name='ChatReaction',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('emoji', models.CharField(max_length=8)),
                ('message', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='reactions', to='chat.chatmessage')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='chat_reactions', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('message', 'user', 'emoji')},
            },
            bases=(models.Model, mojo.models.rest.MojoModel),
        ),
        migrations.CreateModel(
            name='ChatReadReceipt',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('read_at', models.DateTimeField(auto_now_add=True)),
                ('message', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='read_receipts', to='chat.chatmessage')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='chat_read_receipts', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('message', 'user')},
            },
            bases=(models.Model, mojo.models.rest.MojoModel),
        ),
        migrations.AddIndex(
            model_name='chatmessage',
            index=models.Index(fields=['room', 'created'], name='chat_chatme_room_id_36e1c8_idx'),
        ),
        migrations.AddIndex(
            model_name='chatmessage',
            index=models.Index(fields=['room', 'is_flagged', 'created'], name='chat_chatme_room_id_4a80bc_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='chatmembership',
            unique_together={('room', 'user')},
        ),
    ]



"""
TOTP REST endpoints.

Setup flow (authenticated):
  POST /api/account/totp/setup    -> generate secret, return QR code + plain secret
  POST /api/account/totp/confirm  -> verify first code, mark is_enabled=True
  DELETE /api/account/totp        -> disable TOTP

Recovery codes (authenticated):
  GET  /api/account/totp/recovery-codes           -> masked codes + remaining count
  POST /api/account/totp/recovery-codes/regenerate -> new codes (requires valid TOTP code)

Login (2FA verify step — consumes mfa_token):
  POST /api/auth/totp/verify      -> verify code + mfa_token, issue JWT

Recovery login (consumes mfa_token + recovery code):
  POST /api/auth/totp/recover     -> recovery_code + mfa_token, issue JWT

Standalone login (no password):
  POST /api/auth/totp/login       -> username + code -> JWT
"""
from django.db import transaction
from django.db.models import Q

from mojo import decorators as md
from mojo import errors as merrors
from mojo.apps.account.models import User
from mojo.apps.account.models.totp import UserTOTP
from mojo.apps.account.rest.user import jwt_login
from mojo.apps.account.services import mfa as mfa_service
from mojo.apps.account.services import totp as totp_service
from mojo.helpers import logit
from mojo.helpers.qrcode import generate_qrcode
from mojo.helpers.response import JsonResponse


# -----------------------------------------------------------------
# Setup (requires auth)
# -----------------------------------------------------------------

@md.POST("account/totp/setup")
@md.requires_auth()
def on_totp_setup(request):
    """Generate a new TOTP secret and return setup data."""
    secret = totp_service.generate_secret()
    uri = totp_service.get_provisioning_uri(secret, request.user.username)

    # Store unconfirmed secret — not yet enabled
    totp, _ = UserTOTP.objects.get_or_create(user=request.user)
    totp.set_secret("totp_secret", secret)
    totp.is_enabled = False
    totp.save()

    qr = generate_qrcode(data=uri, fmt="base64")

    return JsonResponse({
        "status": True,
        "data": {
            "secret": secret,
            "uri": uri,
            "qr_code": f"data:{qr.content_type};base64,{qr.content}",
        },
    })


@md.POST("account/totp/confirm")
@md.requires_auth()
@md.requires_params("code")
def on_totp_confirm(request):
    """Verify the first TOTP code to activate TOTP for the account."""
    totp = UserTOTP.objects.filter(user=request.user).first()
    if not totp:
        raise merrors.ValueException("TOTP setup not started. Call /api/account/totp/setup first.")

    secret = totp.get_secret("totp_secret")
    if not secret:
        raise merrors.ValueException("TOTP setup not started. Call /api/account/totp/setup first.")

    code = request.DATA.get("code", "").strip()
    if not totp_service.verify_code(secret, code):
        request.user.report_incident("Invalid TOTP confirmation code", "totp:confirm_failed")
        raise merrors.ValueException("Invalid code")

    totp.is_enabled = True
    totp.save()
    codes = totp.generate_recovery_codes()
    request.user.requires_mfa = True
    request.user.save()
    logit.info("account.totp", f"TOTP enabled for user {request.user.username}")
    return JsonResponse({"status": True, "data": {"is_enabled": True, "recovery_codes": codes}})


@md.DELETE("account/totp")
@md.requires_auth()
def on_totp_disable(request):
    """Disable TOTP for the authenticated user."""
    UserTOTP.objects.filter(user=request.user).update(is_enabled=False)
    logit.info("account.totp", f"TOTP disabled for user {request.user.username}")
    return JsonResponse({"status": True})


# -----------------------------------------------------------------
# Recovery codes (requires auth)
# -----------------------------------------------------------------

@md.GET("account/totp/recovery-codes")
@md.requires_auth()
def on_totp_recovery_codes_get(request):
    """Return masked recovery codes and remaining count."""
    totp = UserTOTP.objects.filter(user=request.user, is_enabled=True).first()
    if not totp:
        raise merrors.ValueException("TOTP is not enabled for this account")
    data = totp.get_masked_recovery_codes()
    return JsonResponse({"status": True, "data": data})


@md.POST("account/totp/recovery-codes/regenerate")
@md.requires_auth()
@md.requires_params("code")
def on_totp_recovery_codes_regenerate(request):
    """Regenerate recovery codes. Requires a valid TOTP code to authorize."""
    totp = UserTOTP.objects.filter(user=request.user, is_enabled=True).first()
    if not totp:
        raise merrors.ValueException("TOTP is not enabled for this account")
    secret = totp.get_secret("totp_secret")
    code = request.DATA.get("code", "").strip()
    if not totp_service.verify_code(secret, code):
        raise merrors.PermissionDeniedException("Invalid TOTP code", 403, 403)
    codes = totp.generate_recovery_codes()
    return JsonResponse({"status": True, "data": {"is_enabled": True, "recovery_codes": codes}})


# -----------------------------------------------------------------
# 2FA verify step (consumes mfa_token)
# -----------------------------------------------------------------

@md.POST("auth/totp/verify")
@md.requires_params("mfa_token", "code")
@md.public_endpoint()
def on_totp_verify(request):
    """
    Second factor: verify TOTP code after password login.
    Consumes the mfa_token and issues a full JWT on success.
    """
    token_data = mfa_service.consume_mfa_token(request.DATA.get("mfa_token"))
    if not token_data:
        raise merrors.PermissionDeniedException("Invalid or expired MFA token", 401, 401)

    user = User.objects.filter(pk=token_data["uid"]).first()
    if not user:
        raise merrors.PermissionDeniedException()

    totp = UserTOTP.objects.filter(user=user, is_enabled=True).first()
    if not totp:
        raise merrors.PermissionDeniedException("TOTP not enabled for this account")

    secret = totp.get_secret("totp_secret")
    code = request.DATA.get("code", "").strip()
    if not totp_service.verify_code(secret, code):
        user.report_incident("Invalid TOTP code during login", "totp:login_failed")
        raise merrors.PermissionDeniedException("Invalid code", 401, 401)

    return jwt_login(request, user)


# -----------------------------------------------------------------
# Recovery login (consumes mfa_token + recovery code)
# -----------------------------------------------------------------

@md.POST("auth/totp/recover")
@md.requires_params("mfa_token", "recovery_code")
@md.public_endpoint()
def on_totp_recover(request):
    """
    Authenticate using a recovery code instead of a TOTP code.
    Consumes the mfa_token and one recovery code, then issues a JWT.
    """
    token_data = mfa_service.consume_mfa_token(request.DATA.get("mfa_token"))
    if not token_data:
        raise merrors.PermissionDeniedException("Invalid or expired MFA token", 401, 401)

    user = User.objects.filter(pk=token_data["uid"]).first()
    if not user or not user.is_active:
        raise merrors.PermissionDeniedException()

    with transaction.atomic():
        totp = UserTOTP.objects.filter(user=user, is_enabled=True).select_for_update().first()
        if not totp:
            raise merrors.PermissionDeniedException("TOTP not enabled for this account")

        recovery_code = request.DATA.get("recovery_code", "").strip()
        if not totp.verify_and_consume_recovery_code(recovery_code):
            raise merrors.PermissionDeniedException("Invalid recovery code", 403, 403)

    user.report_incident("TOTP recovery code used", "totp:recovery_used")
    remaining = len(totp.get_secret("recovery_codes") or [])
    if remaining == 0:
        user.notify(
            "You have used your last TOTP recovery code. Generate new ones in your security settings.",
            kind="security",
        )
    return jwt_login(request, user)


# -----------------------------------------------------------------
# Standalone login (username + TOTP code, no password)
# -----------------------------------------------------------------

@md.POST("auth/totp/login")
@md.requires_params("username", "code")
@md.public_endpoint()
def on_totp_login(request):
    """Passwordless login using a TOTP code."""
    username = request.DATA.get("username", "").lower().strip()
    user = User.objects.filter(Q(username=username) | Q(email=username)).first()

    if not user:
        User.class_report_incident(
            f"TOTP login attempt with unknown username: {username}",
            event_type="totp:login_unknown",
            level=8,
            request=request,
        )
        raise merrors.PermissionDeniedException()

    totp = UserTOTP.objects.filter(user=user, is_enabled=True).first()
    if not totp:
        raise merrors.PermissionDeniedException("TOTP not enabled for this account")

    secret = totp.get_secret("totp_secret")
    code = request.DATA.get("code", "").strip()
    if not totp_service.verify_code(secret, code):
        user.report_incident("Invalid TOTP code during standalone login", "totp:login_failed")
        raise merrors.PermissionDeniedException("Invalid code", 401, 401)

    return jwt_login(request, user)



from mojo import errors as merrors


def send_email(to, subject=None, body_text=None, body_html=None,
               cc=None, bcc=None, reply_to=None, mailbox=None, **kwargs):
    """
    Send a plain email using the system default mailbox.

    Args:
        to (str or list): Recipient address or list of addresses.
        subject (str): Email subject line.
        body_text (str): Plain text body (optional but recommended alongside body_html).
        body_html (str): HTML body.
        cc (str or list): CC address or list of addresses.
        bcc (str or list): BCC address or list of addresses.
        reply_to (str or list): Reply-to address or list of addresses.
        mailbox (Mailbox): Override the sending mailbox. Defaults to the system default.
        **kwargs: Additional arguments passed through to the underlying email service.

    Returns:
        SentMessage: The sent message record.

    Raises:
        ValueException: If no mailbox is configured and no override is provided.
        OutboundNotAllowed: If the resolved mailbox has outbound sending disabled.

    Example:
        from mojo.apps import aws

        aws.send_email(
            to="alice@example.com",
            subject="Hello",
            body_html="<p>Hello Alice</p>",
            body_text="Hello Alice"
        )
    """
    from mojo.apps.aws.models import Mailbox

    if mailbox is None:
        mailbox = Mailbox.get_system_default()
    if mailbox is None:
        raise merrors.ValueException("No mailbox configured. Set up a system default mailbox to send email.")

    return mailbox.send_email(
        to=to,
        subject=subject,
        body_text=body_text,
        body_html=body_html,
        cc=cc,
        bcc=bcc,
        reply_to=reply_to,
        **kwargs
    )


def send_template_email(to, template_name, context=None,
                        cc=None, bcc=None, reply_to=None, mailbox=None, **kwargs):
    """
    Send an email using a named EmailTemplate stored in the database.

    The template subject and body are rendered with the supplied context dict using
    {{variable}} syntax. If the resolved mailbox belongs to a domain, a domain-specific
    template override ("{domain_name}.{template_name}") is used automatically when it exists.

    Args:
        to (str or list): Recipient address or list of addresses.
        template_name (str): Name of the EmailTemplate record in the database.
        context (dict): Template variables substituted into the subject and body.
        cc (str or list): CC address or list of addresses.
        bcc (str or list): BCC address or list of addresses.
        reply_to (str or list): Reply-to address or list of addresses.
        mailbox (Mailbox): Override the sending mailbox. Defaults to the system default.
        **kwargs: Additional arguments passed through to the underlying email service.

    Returns:
        SentMessage: The sent message record.

    Raises:
        ValueException: If no mailbox is configured and no override is provided.
        OutboundNotAllowed: If the resolved mailbox has outbound sending disabled.
        ValueError: If the named template does not exist in the database.

    Example:
        from mojo.apps import aws

        aws.send_template_email(
            to="alice@example.com",
            template_name="welcome",
            context={"display_name": "Alice", "app_name": "MyApp"}
        )
    """
    from mojo.apps.aws.models import Mailbox

    if mailbox is None:
        mailbox = Mailbox.get_system_default()
    if mailbox is None:
        raise merrors.ValueException("No mailbox configured. Set up a system default mailbox to send email.")

    return mailbox.send_template_email(
        to=to,
        template_name=template_name,
        context=context,
        cc=cc,
        bcc=bcc,
        reply_to=reply_to,
        **kwargs
    )


def plugin_toc(md):
    """
    A mistune v3 plugin to support a table of contents placeholder [TOC].
    This uses a render hook to perform a simple text substitution.
    """
    def before_render_hook(renderer, text, state):
        # Simple text replacement for the [TOC] placeholder
        return text.replace('[TOC]', '<div class="toc"></div>')

    md.before_render_hooks.append(before_render_hook)

plugin = plugin_toc



from django.db import models
from mojo.models import MojoModel


class Conversation(models.Model, MojoModel):
    """A multi-turn assistant conversation owned by a single user."""

    user = models.ForeignKey("account.User", on_delete=models.CASCADE,
                             related_name="assistant_conversations")
    title = models.CharField(max_length=255, blank=True, default="")
    metadata = models.JSONField(default=dict, blank=True)

    created = models.DateTimeField(auto_now_add=True, editable=False, db_index=True)
    modified = models.DateTimeField(auto_now=True, db_index=True)

    class Meta:
        ordering = ["-modified"]

    class RestMeta:
        NO_REST_SAVE = True
        VIEW_PERMS = ["view_admin", "owner"]
        OWNER_FIELD = "user"
        CAN_DELETE = True
        GRAPHS = {
            "default": {
                "fields": ["id", "title", "created", "modified"],
            },
            "detail": {
                "fields": ["id", "title", "created", "modified", "messages"],
                "graphs": {"messages": "default"},
            },
        }

    def __str__(self):
        return f"Conversation {self.pk} ({self.user})"


class Message(models.Model, MojoModel):
    """A single message in an assistant conversation (user, assistant, or tool)."""

    ROLE_CHOICES = [
        ("user", "User"),
        ("assistant", "Assistant"),
        ("tool_use", "Tool Use"),
        ("tool_result", "Tool Result"),
    ]

    conversation = models.ForeignKey(Conversation, on_delete=models.CASCADE,
                                     related_name="messages")
    role = models.CharField(max_length=16, choices=ROLE_CHOICES, db_index=True)
    content = models.TextField(blank=True, default="")
    tool_calls = models.JSONField(default=None, null=True, blank=True)
    blocks = models.JSONField(default=None, null=True, blank=True)

    created = models.DateTimeField(auto_now_add=True, editable=False, db_index=True)

    class Meta:
        ordering = ["created"]

    class RestMeta:
        NO_REST = True
        VIEW_PERMS = ["view_admin"]
        GRAPHS = {
            "default": {
                "fields": ["id", "role", "content", "tool_calls", "blocks", "created"],
            },
        }

    def __str__(self):
        return f"Message {self.pk} ({self.role})"



from django.apps import AppConfig as BaseAppConfig

class AppConfig(BaseAppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'mojo.apps.account'

    def ready(self):
        from mojo.helpers.settings import settings
        if settings.is_app_installed("django.contrib.admin"):
            self.unregister_apps()

    def unregister_apps(self):
        from django.contrib import admin
        from django.contrib.auth.models import Group
        for model in [Group]:
            admin.site.unregister(model)



import uuid

from mojo.helpers import request as rhelper
import time
from objict import objict
from mojo.helpers.settings import settings
from mojo.helpers import logit
from mojo.models import rest

logger = logit.get_logger("debug", "debug.log")

ANONYMOUS_USER = objict(
    display_name="Anonymous",
    username="anonymous",
    email="anonymous@example.com",
    is_authenticated=False,
    has_permission=lambda: False)

# Cookie TTL for _muid: 2 years in seconds
_MUID_MAX_AGE = 63072000


class MojoMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        request.started_at = time.time()
        request.bearer = None
        request.user = ANONYMOUS_USER
        request.group = None
        request.device = None
        request.request_log = None
        request.ip = rhelper.get_remote_ip(request)
        request.user_agent = rhelper.get_user_agent(request)
        request.duid = rhelper.get_device_id(request)

        # Server-controlled device and session identity (HttpOnly cookies).
        # _muid: persistent device identity (2yr), unforgeable by JS.
        # _msid: browser session identity (no Expires = dies on browser close).
        # These are set on every request so all framework code can use
        # request.muid and request.msid without depending on bouncer.
        request.muid = request.COOKIES.get('_muid', '')
        request.msid = request.COOKIES.get('_msid', '')
        _muid_is_new = not request.muid
        _msid_is_new = not request.msid
        if _muid_is_new:
            request.muid = uuid.uuid4().hex
        if _msid_is_new:
            request.msid = uuid.uuid4().hex

        if settings.LOGIT_REQUEST_BODY:
            request._raw_body = str(request.body)
        else:
            request._raw_body = None
        request.DATA = rhelper.parse_request_data(request)

        # Tab session ID from client JS (sessionStorage, tab-scoped).
        # Only present when mojo-bouncer.js is active on the page.
        request.mtab = request.DATA.get('_mtab', '')

        token = rest.ACTIVE_REQUEST.set(request)
        try:
            resp = self.get_response(request)
            resp.headers["Cache-Control"] = "no-store, no-cache, must-revalidate, max-age=0"
            resp.headers["Pragma"] = "no-cache"
            resp.headers["Expires"] = "0"

            # Set server identity cookies on the response when newly generated.
            if _muid_is_new:
                resp.set_cookie(
                    '_muid', request.muid,
                    max_age=_MUID_MAX_AGE,
                    httponly=True,
                    secure=not settings.DEBUG,
                    samesite='Lax',
                )
            if _msid_is_new:
                # No max_age / expires → session cookie (dies on browser close)
                resp.set_cookie(
                    '_msid', request.msid,
                    httponly=True,
                    secure=not settings.DEBUG,
                    samesite='Lax',
                )

            return resp
        finally:
            rest.ACTIVE_REQUEST.reset(token)



# Generated by Django 5.2.6 on 2025-11-24 00:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incident', '0017_alter_ruleset_bundle_by_alter_ruleset_bundle_minutes_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='ruleset',
            name='bundle_by_rule_set',
            field=models.BooleanField(default=True, help_text='Bundle by rule set'),
        ),
    ]



"""
Tests for the self-service username change endpoint.

Coverage:
  - Happy path — username changes, response contains new username
  - current_password wrong — 401, username unchanged
  - current_password missing — 400
  - username missing — 400
  - username taken by another user — 400
  - username same as current — 400
  - username invalid content (content_guard blocked) — 400
  - username is lowercased on save
  - ALLOW_USERNAME_CHANGE = False — 403 (skip via TestitSkip)
  - Unauthenticated request — 401/403
  - OAuth-only user (no usable password) — 400 with correct message
  - Audit log entry written (username:changed)
"""
from testit import helpers as th
from testit.helpers import assert_true, assert_eq

TEST_USER = "username_change_user"
TEST_PWORD = "uchange##mojo99"
TEST_EMAIL = "username_change_user@example.com"
COLLISION_USER = "username_taken_user"
COLLISION_EMAIL = "username_taken_user@example.com"
OAUTH_USER = "oauth_no_pw_user"
OAUTH_EMAIL = "oauth_no_pw_user@example.com"


# ===========================================================================
# Setup / teardown
# ===========================================================================

@th.django_unit_setup()
def setup_username_change(opts):
    from mojo.apps.account.models import User
    from mojo.decorators.limits import clear_rate_limits
    clear_rate_limits(ip="127.0.0.1")

    # Primary test user
    user = User.objects.filter(email=TEST_EMAIL).last()
    if user is None:
        user = User(username=TEST_USER, email=TEST_EMAIL)
        user.save()
    user.username = TEST_USER
    user.is_active = True
    user.is_email_verified = True
    user.requires_mfa = False
    user.save_password(TEST_PWORD)
    user.save()
    opts.user_id = user.pk

    # Collision user — owns a username so we can test duplicate rejection
    collision = User.objects.filter(email=COLLISION_EMAIL).last()
    if collision is None:
        collision = User(username=COLLISION_USER, email=COLLISION_EMAIL)
        collision.save()
    collision.username = COLLISION_USER
    collision.is_active = True
    collision.is_email_verified = True
    collision.save_password(TEST_PWORD)
    collision.save()
    opts.collision_id = collision.pk

    # OAuth-only user — no usable password
    oauth_user = User.objects.filter(email=OAUTH_EMAIL).last()
    if oauth_user is None:
        oauth_user = User(username=OAUTH_USER, email=OAUTH_EMAIL)
        oauth_user.set_unusable_password()
        oauth_user.is_email_verified = True
        oauth_user.save()
    else:
        oauth_user.username = OAUTH_USER
        oauth_user.set_unusable_password()
        oauth_user.is_active = True
        oauth_user.is_email_verified = True
        oauth_user.save()
    opts.oauth_user_id = oauth_user.pk


# ===========================================================================
# Endpoint tests
# ===========================================================================

@th.django_unit_test("username change: happy path")
def test_username_change_happy(opts):
    from mojo.apps.account.models import User

    opts.client.login(TEST_USER, TEST_PWORD)
    resp = opts.client.post("/api/auth/username/change", {
        "username": "new_uname_test",
        "current_password": TEST_PWORD,
    })
    opts.client.logout()
    assert_eq(resp.status_code, 200, f"Expected 200, got {resp.status_code}")
    data = resp.json
    assert_true(data.get("status"), "Expected status=true")
    assert_eq(data.get("data", {}).get("username"), "new_uname_test", "Response should contain new username")

    # Verify persisted
    user = User.objects.get(pk=opts.user_id)
    assert_eq(user.username, "new_uname_test", "Username should be persisted in DB")

    # Restore original username for subsequent tests
    user.username = TEST_USER
    user.save(update_fields=["username", "modified"])


@th.django_unit_test("username change: wrong password returns 401")
def test_username_change_wrong_password(opts):
    from mojo.apps.account.models import User

    opts.client.login(TEST_USER, TEST_PWORD)
    resp = opts.client.post("/api/auth/username/change", {
        "username": "should_not_change",
        "current_password": "wrong_password_here",
    })
    opts.client.logout()
    assert_eq(resp.status_code, 401, f"Expected 401, got {resp.status_code}")

    # Username must not have changed
    user = User.objects.get(pk=opts.user_id)
    assert_eq(user.username, TEST_USER, "Username should be unchanged after wrong password")


@th.django_unit_test("username change: missing current_password returns 400")
def test_username_change_missing_password(opts):
    opts.client.login(TEST_USER, TEST_PWORD)
    resp = opts.client.post("/api/auth/username/change", {
        "username": "new_uname",
    })
    opts.client.logout()
    assert_true(resp.status_code in (400, 422), f"Expected 400, got {resp.status_code}")


@th.django_unit_test("username change: missing username returns 400")
def test_username_change_missing_username(opts):
    opts.client.login(TEST_USER, TEST_PWORD)
    resp = opts.client.post("/api/auth/username/change", {
        "current_password": TEST_PWORD,
    })
    opts.client.logout()
    assert_true(resp.status_code in (400, 422), f"Expected 400, got {resp.status_code}")


@th.django_unit_test("username change: taken by another user returns 400")
def test_username_change_taken(opts):
    opts.client.login(TEST_USER, TEST_PWORD)
    resp = opts.client.post("/api/auth/username/change", {
        "username": COLLISION_USER,
        "current_password": TEST_PWORD,
    })
    opts.client.logout()
    assert_true(resp.status_code in (400, 422), f"Expected 400, got {resp.status_code}")


@th.django_unit_test("username change: same as current returns 400")
def test_username_change_same(opts):
    opts.client.login(TEST_USER, TEST_PWORD)
    resp = opts.client.post("/api/auth/username/change", {
        "username": TEST_USER,
        "current_password": TEST_PWORD,
    })
    opts.client.logout()
    assert_true(resp.status_code in (400, 422), f"Expected 400, got {resp.status_code}")


@th.django_unit_test("username change: username is lowercased on save")
def test_username_change_lowercase(opts):
    from mojo.apps.account.models import User

    opts.client.login(TEST_USER, TEST_PWORD)
    resp = opts.client.post("/api/auth/username/change", {
        "username": "MiXeD_CaSe_NaMe",
        "current_password": TEST_PWORD,
    })
    opts.client.logout()
    assert_eq(resp.status_code, 200, f"Expected 200, got {resp.status_code}")
    data = resp.json
    assert_eq(data.get("data", {}).get("username"), "mixed_case_name",
              "Username should be lowercased")

    user = User.objects.get(pk=opts.user_id)
    assert_eq(user.username, "mixed_case_name", "DB username should be lowercased")

    # Restore
    user.username = TEST_USER
    user.save(update_fields=["username", "modified"])


@th.django_unit_test("username change: unauthenticated returns 401/403")
def test_username_change_unauth(opts):
    opts.client.logout()
    resp = opts.client.post("/api/auth/username/change", {
        "username": "no_auth_user",
        "current_password": TEST_PWORD,
    })
    assert_true(resp.status_code in (401, 403), f"Expected 401 or 403, got {resp.status_code}")


@th.django_unit_test("username change: OAuth-only user (no password) returns 400")
def test_username_change_oauth_no_password(opts):
    from mojo.apps.account.models import User

    # Give oauth user a temp password so we can log in, then remove it
    oauth_user = User.objects.get(pk=opts.oauth_user_id)
    oauth_user.save_password("temp_pass_1234")
    oauth_user.save()

    opts.client.login(OAUTH_USER, "temp_pass_1234")

    # Now remove the usable password while we still have a valid session.
    # Use update_fields to avoid overwriting server-side state (last_login,
    # last_activity) that was modified by the login call above.
    oauth_user.set_unusable_password()
    oauth_user.save(update_fields=["password", "modified"])

    resp = opts.client.post("/api/auth/username/change", {
        "username": "new_oauth_name",
        "current_password": "anything",
    })
    opts.client.logout()
    assert_true(resp.status_code in (400, 422), f"Expected 400, got {resp.status_code}")
    body = resp.json
    assert_true("password" in str(body).lower() or "No password" in str(body),
                "Error message should mention password")


@th.django_unit_test("username change: ALLOW_USERNAME_CHANGE=False returns 403")
def test_username_change_disabled(opts):
    from mojo.helpers.settings import settings
    from testit import TestitSkip

    if settings.get("ALLOW_USERNAME_CHANGE", True):
        raise TestitSkip("ALLOW_USERNAME_CHANGE is True on this server — cannot test disabled state")

    opts.client.login(TEST_USER, TEST_PWORD)
    resp = opts.client.post("/api/auth/username/change", {
        "username": "disabled_change",
        "current_password": TEST_PWORD,
    })
    opts.client.logout()
    assert_eq(resp.status_code, 403, f"Expected 403, got {resp.status_code}")


@th.django_unit_test("username change: audit log entry written")
def test_username_change_audit_log(opts):
    from mojo.apps.account.models import User
    from mojo.apps.incident.models.event import Event

    opts.client.login(TEST_USER, TEST_PWORD)

    # Record event count before
    before_count = Event.objects.filter(
        uid=opts.user_id, category="username:changed"
    ).count()

    resp = opts.client.post("/api/auth/username/change", {
        "username": "audit_log_uname",
        "current_password": TEST_PWORD,
    })
    opts.client.logout()
    assert_eq(resp.status_code, 200, f"Expected 200, got {resp.status_code}")

    # Restore username
    user = User.objects.get(pk=opts.user_id)
    user.username = TEST_USER
    user.save(update_fields=["username", "modified"])


"""Metrics domain tools — fetch metrics, system health, incident trends."""
import re
from mojo.helpers import dates

# Valid account scopes for metrics
_VALID_ACCOUNT_RE = re.compile(r"^(public|global|group-\d+|user-\d+)$")


def _tool_fetch_metrics(params, user):
    from mojo.apps import metrics

    slugs = params.get("slugs", [])
    if not slugs:
        return {"error": "At least one slug is required"}

    granularity = params.get("granularity", "hours")
    account = params.get("account", "public")
    if not _VALID_ACCOUNT_RE.match(account):
        return {"error": f"Invalid account scope: {account}. Use public, global, group-<id>, or user-<id>."}
    dt_start = None
    dt_end = None

    if params.get("dt_start"):
        dt_start = dates.parse(params["dt_start"])
    if params.get("dt_end"):
        dt_end = dates.parse(params["dt_end"])

    if len(slugs) == 1:
        slugs = slugs[0]

    records = metrics.fetch(
        slugs, dt_start=dt_start, dt_end=dt_end,
        granularity=granularity, account=account,
        with_labels=True, allow_empty=True,
    )
    return records


def _tool_get_system_health(params, user):
    """Aggregate cross-domain health stats."""
    from mojo.apps.account.models import User
    from mojo.apps.incident.models import Incident, Event
    from mojo.apps.jobs.models import Job
    from django.db.models import Count

    now_minus_1h = dates.subtract(minutes=60)
    now_minus_24h = dates.subtract(minutes=1440)

    # Active users (last_activity in last hour)
    active_users = User.objects.filter(
        last_activity__gte=now_minus_1h, is_active=True
    ).count()

    # Open incidents
    open_incidents = Incident.objects.filter(
        status__in=["new", "open", "investigating"]
    ).count()

    # Events in last hour
    events_1h = Event.objects.filter(created__gte=now_minus_1h).count()

    # Job queue
    pending_jobs = Job.objects.filter(status="pending").count()
    running_jobs = Job.objects.filter(status="running").count()
    failed_24h = Job.objects.filter(
        status="failed", created__gte=now_minus_24h
    ).count()

    return {
        "active_users_1h": active_users,
        "open_incidents": open_incidents,
        "events_last_hour": events_1h,
        "pending_jobs": pending_jobs,
        "running_jobs": running_jobs,
        "failed_jobs_24h": failed_24h,
    }


def _tool_get_incident_trends(params, user):
    """Incident and event counts over recent time periods for comparison."""
    from mojo.apps.incident.models import Incident, Event

    periods = [
        ("last_1h", 60),
        ("last_6h", 360),
        ("last_24h", 1440),
        ("last_7d", 10080),
    ]

    result = {}
    for label, minutes in periods:
        since = dates.subtract(minutes=minutes)
        result[label] = {
            "incidents": Incident.objects.filter(created__gte=since).count(),
            "events": Event.objects.filter(created__gte=since).count(),
        }

    # Category breakdown for last 24h
    from django.db.models import Count
    since_24h = dates.subtract(minutes=1440)
    categories = list(
        Event.objects.filter(created__gte=since_24h)
        .values("category")
        .annotate(count=Count("id"))
        .order_by("-count")[:20]
    )
    result["categories_24h"] = categories

    return result


# ---------------------------------------------------------------------------
# Tool definitions
# ---------------------------------------------------------------------------

TOOLS = [
    {
        "name": "fetch_metrics",
        "description": "Fetch time-series metrics data for given slugs and date range.",
        "input_schema": {
            "type": "object",
            "properties": {
                "slugs": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "List of metric slugs to fetch",
                },
                "dt_start": {"type": "string", "description": "Start date (ISO format, optional)"},
                "dt_end": {"type": "string", "description": "End date (ISO format, optional)"},
                "granularity": {"type": "string", "description": "Data granularity (hours, days, months)", "default": "hours"},
                "account": {"type": "string", "description": "Account scope (public, global, group-<id>)", "default": "public"},
            },
            "required": ["slugs"],
        },
        "handler": _tool_fetch_metrics,
        "permission": "view_admin",
    },
    {
        "name": "get_system_health",
        "description": "Overview of system health: active users, job queue depth, error rates, open incident counts.",
        "input_schema": {
            "type": "object",
            "properties": {},
        },
        "handler": _tool_get_system_health,
        "permission": "view_admin",
    },
    {
        "name": "get_incident_trends",
        "description": "Incident and event trends over time (1h, 6h, 24h, 7d) with category breakdown.",
        "input_schema": {
            "type": "object",
            "properties": {},
        },
        "handler": _tool_get_incident_trends,
        "permission": "view_security",
    },
]



from testit import helpers as th


@th.django_unit_test()
def test_domain_convenience_functions(opts):
    """Test that convenience functions exist and are callable."""
    from mojo.helpers import domain

    # All these should be callable
    assert callable(domain.a), "Expected domain.a to be callable"
    assert callable(domain.ip), "Expected domain.ip to be callable"
    assert callable(domain.ips), "Expected domain.ips to be callable"
    assert callable(domain.mx), "Expected domain.mx to be callable"
    assert callable(domain.txt), "Expected domain.txt to be callable"
    assert callable(domain.reverse), "Expected domain.reverse to be callable"
    assert callable(domain.dns), "Expected domain.dns to be callable"
    assert callable(domain.lookup), "Expected domain.lookup to be callable"
    assert callable(domain.spf), "Expected domain.spf to be callable"
    assert callable(domain.dmarc), "Expected domain.dmarc to be callable"
    assert callable(domain.dkim), "Expected domain.dkim to be callable"
    assert callable(domain.whois), "Expected domain.whois to be callable"
    assert callable(domain.is_available), "Expected domain.is_available to be callable"
    assert callable(domain.ssl), "Expected domain.ssl to be callable"


@th.django_unit_test()
def test_domain_ip_single(opts):
    """Test single IP lookup."""
    from mojo.helpers import domain

    # Test well-known domain - default (first IP)
    ip = domain.ip('google.com')
    assert isinstance(ip, str), "Expected string IP"
    assert len(ip) > 0, "Expected IP for google.com"
    assert '.' in ip, "Expected IPv4 address"

    # Test random selection
    ip_random = domain.ip('google.com', random=True)
    assert isinstance(ip_random, str), "Expected string IP"
    assert '.' in ip_random, "Expected IPv4 address"

    # Test subdomain
    ip = domain.ip('mail.google.com')
    assert isinstance(ip, str), "Expected string IP"

    # Test invalid domain - should return empty string
    ip = domain.ip('thisdoesnotexist12345xyz.com')
    assert ip == '', "Expected empty string for non-existent domain"


@th.django_unit_test()
def test_domain_ips_simple(opts):
    """Test simple IP lookup."""
    from mojo.helpers import domain

    # Test well-known domain
    ips = domain.ips('google.com')
    assert isinstance(ips, list), "Expected list of IPs"
    assert len(ips) > 0, "Expected at least one IP for google.com"
    assert all('.' in ip for ip in ips), "Expected IPv4 addresses"

    # Test subdomain
    ips = domain.ips('mail.google.com')
    assert isinstance(ips, list), "Expected list of IPs"

    # Test invalid domain - should return empty list
    ips = domain.ips('thisdoesnotexist12345xyz.com')
    assert ips == [], "Expected empty list for non-existent domain"


@th.django_unit_test()
def test_domain_a_lookup(opts):
    """Test A record lookup."""
    from mojo.helpers import domain

    # Test well-known domain
    result = domain.a('google.com')
    assert result.records, "Expected A records for google.com"
    assert not result.error, f"Expected no error, got {result.error}"
    assert all('.' in ip for ip in result.records), "Expected IPv4 addresses"

    # Test invalid domain
    result = domain.a('thisdoesnotexist12345xyz.com')
    assert result.error, "Expected error for non-existent domain"
    assert not result.records, "Expected no records for non-existent domain"


@th.django_unit_test()
def test_domain_mx_lookup(opts):
    """Test MX record lookup."""
    from mojo.helpers import domain

    # Test well-known domain
    result = domain.mx('google.com')
    assert result.records, "Expected MX records for google.com"
    assert not result.error, f"Expected no error, got {result.error}"

    # Verify MX record structure
    mx = result.records[0]
    assert 'priority' in mx, "Expected priority in MX record"
    assert 'host' in mx, "Expected host in MX record"
    assert isinstance(mx['priority'], int), "Expected priority to be int"
    assert isinstance(mx['host'], str), "Expected host to be string"

    # Verify sorted by priority
    if len(result.records) > 1:
        priorities = [mx['priority'] for mx in result.records]
        assert priorities == sorted(priorities), "Expected MX records sorted by priority"


@th.django_unit_test()
def test_domain_txt_lookup(opts):
    """Test TXT record lookup."""
    from mojo.helpers import domain

    # Test well-known domain
    result = domain.txt('google.com')
    assert isinstance(result.records, list), "Expected records list"
    assert not result.error or 'No TXT' in result.error, f"Expected no error or no TXT records"


@th.django_unit_test()
def test_domain_reverse_lookup(opts):
    """Test reverse DNS lookup."""
    from mojo.helpers import domain

    # Test Google DNS
    result = domain.reverse('8.8.8.8')
    assert result.hostname, "Expected hostname for 8.8.8.8"
    assert not result.error, f"Expected no error, got {result.error}"
    assert 'google' in result.hostname.lower(), f"Expected 'google' in hostname, got {result.hostname}"


@th.django_unit_test()
def test_domain_dns_query(opts):
    """Test generic DNS query."""
    from mojo.helpers import domain

    # Test AAAA records (IPv6)
    result = domain.dns('google.com', 'AAAA')
    assert isinstance(result.records, list), "Expected records list"

    # Test NS records
    result = domain.dns('google.com', 'NS')
    assert result.records, "Expected NS records for google.com"
    assert not result.error, f"Expected no error, got {result.error}"


@th.django_unit_test()
def test_domain_lookup_comprehensive(opts):
    """Test comprehensive domain lookup."""
    from mojo.helpers import domain

    result = domain.lookup('google.com')
    assert result.domain == 'google.com', f"Expected domain to be google.com, got {result.domain}"
    assert result.a, "Expected A records"
    assert result.mx, "Expected MX records"
    assert isinstance(result.txt, list), "Expected TXT records list"
    assert not result.error, f"Expected no error, got {result.error}"

    # Verify structure
    assert all('.' in ip for ip in result.a), "Expected IPv4 addresses"
    assert all('priority' in mx and 'host' in mx for mx in result.mx), "Expected MX record structure"


@th.django_unit_test()
def test_domain_spf_lookup(opts):
    """Test SPF record lookup and parsing."""
    from mojo.helpers import domain

    # Test domain with SPF
    result = domain.spf('google.com')

    if result.valid:
        assert result.record, "Expected SPF record"
        assert result.record.startswith('v=spf1'), "Expected SPF record to start with v=spf1"
        assert result.parsed, "Expected parsed SPF data"
        assert result.parsed.version == 'spf1', f"Expected version spf1, got {result.parsed.version}"
        assert result.parsed.mechanisms, "Expected SPF mechanisms"
        assert result.parsed.qualifier, "Expected SPF qualifier"
    else:
        # It's okay if no SPF record found
        assert result.error or not result.record, "Expected error or no record"


@th.django_unit_test()
def test_domain_dmarc_lookup(opts):
    """Test DMARC record lookup and parsing."""
    from mojo.helpers import domain

    # Test domain with DMARC
    result = domain.dmarc('google.com')

    if result.valid:
        assert result.record, "Expected DMARC record"
        assert result.record.startswith('v=DMARC1'), "Expected DMARC record to start with v=DMARC1"
        assert result.parsed, "Expected parsed DMARC data"
        assert result.parsed.version == 'DMARC1', f"Expected version DMARC1"
        assert result.parsed.p, "Expected DMARC policy"
    else:
        # It's okay if no DMARC record found
        assert result.error or not result.record, "Expected error or no record"


@th.django_unit_test()
def test_domain_dkim_lookup(opts):
    """Test DKIM record lookup and parsing."""
    from mojo.helpers import domain

    # Test with common selector
    result = domain.dkim('google.com', selector='google')

    # DKIM may or may not be found depending on selector
    if result.valid:
        assert result.record, "Expected DKIM record"
        assert result.parsed, "Expected parsed DKIM data"
        assert result.parsed.p or result.parsed.public_key, "Expected public key in DKIM record"

    # Test invalid selector
    result = domain.dkim('google.com', selector='thisdoesnotexist12345')
    assert result.error or not result.valid, "Expected error or invalid for non-existent selector"


@th.django_unit_test()
def test_domain_whois_lookup(opts):
    """Test WHOIS lookup."""
    from mojo.helpers import domain

    # Test well-known domain
    result = domain.whois('google.com')

    if not result.error:
        assert result.domain_name, "Expected domain_name in WHOIS result"
        assert 'google.com' in result.domain_name.lower(), f"Expected google.com in domain_name"

        # Check for common WHOIS fields (may vary by registrar)
        # At least some of these should be present
        has_data = any([
            result.get('registrar'),
            result.get('creation_date'),
            result.get('expiration_date'),
            result.get('name_servers')
        ])
        assert has_data, "Expected at least some WHOIS data"
    else:
        # WHOIS may be rate-limited or blocked
        assert result.error, "Expected error message if WHOIS failed"


@th.django_unit_test()
def test_domain_is_available(opts):
    """Test domain availability check."""
    from mojo.helpers import domain

    # Test registered domain
    result = domain.is_available('google.com')
    assert result.domain == 'google.com', f"Expected domain to be google.com"
    assert result.available == False, "Expected google.com to not be available"
    assert result.reason, "Expected reason for availability status"

    # Test likely unregistered domain (very long random string)
    import random
    import string
    random_domain = ''.join(random.choices(string.ascii_lowercase + string.digits, k=63)) + '.com'
    result = domain.is_available(random_domain)
    assert result.domain == random_domain, f"Expected domain to be {random_domain}"
    # May be available or may get an error


@th.django_unit_test()
def test_domain_ssl_certificate(opts):
    """Test SSL certificate lookup."""
    from mojo.helpers import domain

    # Test well-known domain with valid SSL
    result = domain.ssl('google.com')

    if not result.error:
        assert result.subject, "Expected subject in SSL certificate"
        assert result.issuer, "Expected issuer in SSL certificate"
        assert result.valid_from, "Expected valid_from date"
        assert result.valid_until, "Expected valid_until date"
        assert result.days_remaining is not None, "Expected days_remaining"
        assert result.expired is not None, "Expected expired flag"
        assert isinstance(result.san, list), "Expected SAN to be list"

        # Verify certificate is not expired for google.com
        assert not result.expired, "Expected google.com certificate to not be expired"
        assert result.days_remaining > 0, "Expected positive days remaining for google.com"

        # Verify issuer info
        assert result.issuer, "Expected issuer information"
    else:
        # SSL lookup may fail due to network issues
        assert result.error, "Expected error message if SSL lookup failed"


@th.django_unit_test()
def test_domain_objict_access(opts):
    """Test that all results return objict instances."""
    from mojo.helpers import domain
    from objict import objict

    # Test A record
    result = domain.a('google.com')
    assert isinstance(result, objict), "Expected objict instance"
    assert hasattr(result, 'records'), "Expected records attribute"
    assert hasattr(result, 'error'), "Expected error attribute"

    # Test MX record
    result = domain.mx('google.com')
    assert isinstance(result, objict), "Expected objict instance"

    # Test comprehensive lookup
    result = domain.lookup('google.com')
    assert isinstance(result, objict), "Expected objict instance"
    assert result.domain == 'google.com', "Expected domain attribute access"

    # Test WHOIS
    result = domain.whois('google.com')
    assert isinstance(result, objict), "Expected objict instance"

    # Test SSL
    result = domain.ssl('google.com')
    assert isinstance(result, objict), "Expected objict instance"


@th.django_unit_test()
def test_domain_error_handling(opts):
    """Test error handling for invalid inputs."""
    from mojo.helpers import domain

    # Test invalid domain name
    result = domain.a('invalid..domain..com')
    assert result.error or not result.records, "Expected error or no records for invalid domain"

    # Test non-existent domain
    result = domain.a('thisdoesnotexist12345xyz.com')
    assert result.error, "Expected error for non-existent domain"

    # Test invalid IP for reverse lookup
    result = domain.reverse('999.999.999.999')
    assert result.error, "Expected error for invalid IP"

    # All errors should have error field
    assert hasattr(result, 'error'), "Expected error attribute"
    assert result.error, "Expected error message"


@th.django_unit_test()
def test_domain_normalization(opts):
    """Test domain name normalization."""
    from mojo.helpers import domain

    # Test uppercase domain
    result1 = domain.a('GOOGLE.COM')
    result2 = domain.a('google.com')

    # Should both work (normalization happens internally)
    if not result1.error and not result2.error:
        assert result1.records, "Expected records for uppercase domain"
        assert result2.records, "Expected records for lowercase domain"

    # Test with spaces
    result = domain.a('  google.com  ')
    if not result.error:
        assert result.records, "Expected records for domain with spaces"



# Generated by Django 4.2.21 on 2025-08-31 23:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aws', '0002_emaildomain_can_recv_emaildomain_can_send_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='mailbox',
            name='is_domain_default',
            field=models.BooleanField(db_index=True, default=False, help_text='Default mailbox for this domain (one per domain)'),
        ),
        migrations.AddField(
            model_name='mailbox',
            name='is_system_default',
            field=models.BooleanField(db_index=True, default=False, help_text='System-wide default mailbox (only one allowed)'),
        ),
        migrations.AddIndex(
            model_name='mailbox',
            index=models.Index(fields=['is_system_default'], name='aws_mailbox_is_syst_9911c7_idx'),
        ),
        migrations.AddIndex(
            model_name='mailbox',
            index=models.Index(fields=['is_domain_default', 'domain'], name='aws_mailbox_is_doma_189efb_idx'),
        ),
    ]



# Generated by Django 4.2.21 on 2025-08-30 21:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0013_pushconfig_test_mode_alter_pushconfig_apns_enabled_and_more'),
        ('docit', '0002_alter_book_modified_by_alter_page_modified_by'),
    ]

    operations = [
        migrations.AlterField(
            model_name='book',
            name='group',
            field=models.ForeignKey(blank=True, default=None, help_text='Owning group for this book', null=True, on_delete=django.db.models.deletion.CASCADE, to='account.group'),
        ),
    ]



from django.db import models
from mojo.models import MojoModel
from mojo.helpers import dates, logit


KIND_CHOICES = [
    ("country", "Country"),
    ("datacenter", "Datacenter"),
    ("abuse", "Abuse List"),
    ("custom", "Custom"),
]

SOURCE_CHOICES = [
    ("ipdeny", "ipdeny.com"),
    ("abuseipdb", "AbuseIPDB"),
    ("manual", "Manual"),
]


class IPSet(models.Model, MojoModel):
    """
    Manages ipset-based bulk IP blocking (countries, datacenters, abuse lists).

    Each record represents one ipset (e.g. "country_cn", "abuse_ips", "azure").
    The CIDR data is stored directly in the `data` TextField — no external
    file dependency at sync time.

    The sync action broadcasts to all instances so every EC2 behind the
    load balancer gets the same ipset rules.
    """
    created = models.DateTimeField(auto_now_add=True, editable=False, db_index=True)
    modified = models.DateTimeField(auto_now=True, db_index=True)

    name = models.CharField(max_length=100, unique=True, help_text="ipset name, e.g. country_cn, abuse_ips, azure")
    kind = models.CharField(max_length=20, choices=KIND_CHOICES, db_index=True)
    description = models.CharField(max_length=255, null=True, blank=True)

    # Where the data comes from (for auto-refresh)
    source = models.CharField(max_length=20, choices=SOURCE_CHOICES, default="manual")
    source_url = models.CharField(max_length=500, null=True, blank=True, help_text="URL to fetch CIDR list from (ipdeny, etc.)")
    source_key = models.CharField(max_length=255, null=True, blank=True, help_text="API key or identifier for the source")

    # The actual CIDR data — one CIDR per line
    data = models.TextField(default="", blank=True, help_text="CIDR list, one per line")

    is_enabled = models.BooleanField(default=True, db_index=True)
    cidr_count = models.IntegerField(default=0)
    last_synced = models.DateTimeField(null=True, blank=True)
    sync_error = models.TextField(null=True, blank=True)

    class Meta:
        verbose_name = "IP Set"
        verbose_name_plural = "IP Sets"

    class RestMeta:
        VIEW_PERMS = ["view_security", "security"]
        SAVE_PERMS = ["manage_security", "security"]
        DELETE_PERMS = ["manage_security"]
        SEARCH_FIELDS = ["name", "description"]
        POST_SAVE_ACTIONS = ["sync", "enable", "disable", "refresh_source"]
        CAN_DELETE = True
        GRAPHS = {
            "default": {
                "exclude": ["data", "source_key"],
            },
            "detailed": {
                "exclude": ["source_key"],
            },
        }

    def __str__(self):
        status = "enabled" if self.is_enabled else "disabled"
        return f"{self.name} ({self.kind}, {self.cidr_count} CIDRs, {status})"

    @property
    def cidrs(self):
        """Returns the CIDR list as a list of strings."""
        if not self.data:
            return []
        return [line.strip() for line in self.data.strip().splitlines() if line.strip() and not line.startswith("#")]

    def set_data(self, cidr_list):
        """Set CIDR data from a list of strings."""
        self.data = "\n".join(cidr_list)
        self.cidr_count = len(cidr_list)

    def on_action_sync(self, value):
        """Broadcast ipset to all instances."""
        self.sync()

    def on_action_enable(self, value):
        self.is_enabled = True
        self.save(update_fields=["is_enabled"])
        self.sync()

    def on_action_disable(self, value):
        """Disable and remove ipset from all instances."""
        self.is_enabled = False
        self.save(update_fields=["is_enabled"])
        from mojo.apps import jobs
        jobs.broadcast_execute(
            "mojo.apps.incident.asyncjobs.broadcast_remove_ipset",
            {"name": self.name},
        )

    def on_action_refresh_source(self, value):
        """Fetch latest data from source, then sync to all instances."""
        if self.refresh_from_source():
            self.sync()

    def sync(self):
        """Broadcast this ipset to all instances."""
        if not self.is_enabled:
            return
        from mojo.apps import jobs
        jobs.broadcast_execute(
            "mojo.apps.incident.asyncjobs.broadcast_sync_ipset",
            {"name": self.name, "cidrs": self.cidrs},
        )
        self.last_synced = dates.utcnow()
        self.sync_error = None
        self.save(update_fields=["last_synced", "sync_error"])

    def refresh(self):
        """Refresh the ipset from the source and sync."""
        if self.refresh_from_source():
            self.sync()

    def refresh_from_source(self):
        """Fetch latest CIDR data from the configured source."""
        if self.source == "manual":
            return False

        try:
            if self.source == "ipdeny":
                data = self._fetch_ipdeny()
            elif self.source == "abuseipdb":
                data = self._fetch_abuseipdb()
            else:
                return False

            if data:
                self.set_data(data)
                self.sync_error = None
                self.save(update_fields=["data", "cidr_count", "sync_error"])
                return True
            return False
        except Exception as e:
            self.sync_error = str(e)
            self.save(update_fields=["sync_error"])
            logit.error(f"IPSet refresh failed for {self.name}: {e}")
            return False

    def _fetch_ipdeny(self):
        """Fetch country zone file from ipdeny.com."""
        import requests
        if not self.source_url:
            if not self.name or not self.name.startswith("country_"):
                raise ValueError(
                    f"IPSet '{self.name}' has source=ipdeny but no source_url and "
                    f"name does not start with 'country_' — cannot derive URL"
                )
            import re
            code = self.name[len("country_"):]
            if not re.fullmatch(r'[a-z]{2}', code):
                raise ValueError(
                    f"IPSet '{self.name}': derived country code '{code}' is not a valid "
                    f"2-letter code — cannot construct ipdeny URL"
                )
            self.source_url = f"https://www.ipdeny.com/ipblocks/data/countries/{code}.zone"
            self.save(update_fields=["source_url"])
        resp = requests.get(self.source_url, timeout=30)
        resp.raise_for_status()
        lines = [line.strip() for line in resp.text.splitlines() if line.strip() and not line.startswith("#")]
        return lines

    def _fetch_abuseipdb(self):
        """Fetch abuse IP blacklist from AbuseIPDB API."""
        import requests
        if not self.source_key:
            return None
        resp = requests.get(
            "https://api.abuseipdb.com/api/v2/blacklist",
            headers={"Key": self.source_key, "Accept": "text/plain"},
            params={"confidenceMinimum": 100, "limit": 10000, "ipVersion": 4, "plaintext": ""},
            timeout=30,
        )
        resp.raise_for_status()
        lines = [line.strip() for line in resp.text.splitlines() if line.strip()]
        return lines

    @classmethod
    def create_country(cls, country_code, enabled=True):
        """Helper to create a country IPSet."""
        code = country_code.lower()
        return cls.objects.update_or_create(
            name=f"country_{code}",
            defaults={
                "kind": "country",
                "description": f"Block country: {code.upper()}",
                "source": "ipdeny",
                "source_url": f"https://www.ipdeny.com/ipblocks/data/countries/{code}.zone",
                "is_enabled": enabled,
            }
        )[0]

    @classmethod
    def create_abuse_list(cls, api_key, enabled=True):
        """Helper to create an AbuseIPDB IPSet."""
        return cls.objects.update_or_create(
            name="abuse_ips",
            defaults={
                "kind": "abuse",
                "description": "AbuseIPDB blacklist (confidence 100%)",
                "source": "abuseipdb",
                "source_key": api_key,
                "is_enabled": enabled,
            }
        )[0]



"""
Redis adapter for the jobs system.
Imports the framework-level RedisAdapter with backward compatibility.
"""
# Import from framework
from mojo.helpers.redis import RedisAdapter, get_adapter as get_framework_adapter, reset_adapter


# Maintain backward compatibility for jobs module
def get_adapter() -> RedisAdapter:
    """
    Get the default Redis adapter instance for jobs.

    Returns:
        RedisAdapter instance from framework
    """
    return get_framework_adapter()


# Expose reset function for testing
def reset_adapter():
    """Reset the default adapter (useful for testing)."""
    from mojo.helpers.redis import reset_adapter as framework_reset
    framework_reset()
